from typing import Any, Self
import datetime as dt
import pandas as pd
from pandas import Timestamp, Timedelta

Time = str | dt.date | dt.datetime | Timestamp
Delta = dt.timedelta | Timedelta


class FlexiTime:
    """Flexible wrapper around python's own ``datetime.datetime`` object.

    Parameters
    ----------
    time
        Can be an ISO string of a date or datetime, a ``datetime.date`` or
        a ``datetime.datetime`` object, or a ``pandas.Timestamp``.

    """

    def __init__(self, time: Time | Self) -> None:
        self.as_datetime = dt.datetime.fromisoformat(str(time))

    def __getattr__(self, attribute: str) -> Any:
        return getattr(self.as_datetime, attribute)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{str(self)}')"

    def __str__(self) -> str:
        return str(self.as_datetime)

    def __lt__(self, other: Time) -> bool:
        return self.as_datetime < self.__class__(other).as_datetime

    def __le__(self, other: Time) -> bool:
        return self.as_datetime <= self.__class__(other).as_datetime

    def __gt__(self, other: Time) -> bool:
        return self.as_datetime > self.__class__(other).as_datetime

    def __ge__(self, other: Time) -> bool:
        return self.as_datetime >= self.__class__(other).as_datetime

    def __eq__(self, other: Time) -> bool:
        return self.__class__(other).as_datetime == self.as_datetime

    def __ne__(self, other: Time) -> bool:
        return self.__class__(other).as_datetime != self.as_datetime

    def __add__(self, other: Delta) -> Self:
        return self.__class__(self.as_datetime + other)

    def __radd__(self, other: Delta) -> Self:
        return self.__class__(other + self.as_datetime)

    def __sub__(self, other: Delta) -> Self:
        return self.__class__(self.as_datetime - other)

    def __hash__(self) -> int:
        return self.as_datetime.__hash__()

    @property
    def as_json(self) -> str:
        """Representation to appear in a JSON."""
        return str(self)

    @property
    def as_dtype(self) -> Timestamp:
        """Representation in a pandas DataFrame."""
        return pd.to_datetime(self.as_datetime)



"""Convenient partials of functions for manipulating dictionaries.

Parameters that are known at program start are used to initialize the classes
so that, at runtime, dictionaries can flow through a preconfigured processing
pipe of callable objects.

"""

from .valuesgetter import ValuesGetter

__all__ = [
    'ValuesGetter'
]



"""Tools to assist in consolidating sources of project configurations."""

from .importer import Importer
from .envparser import EnvParser
from .argparser import ArgParser, USAGE, DESCRIPTION, EPILOG

__all__ = [
    'Importer',
    'EnvParser',
    'ArgParser',
    'USAGE',
    'DESCRIPTION',
    'EPILOG'
]



from typing import Any
import yaml
from yaml import Loader
from ..misc import ArgRepr

type Yaml = dict[str, Any] | list[Any]


class YamlParser(ArgRepr):
    """Light wrapper around pyyaml's ``yaml.load`` function.

    Parameters
    ----------
    loader: type, optional
        The loader class to use. Defaults to ``Loader``

    """

    def __init__(self, loader: type = Loader) -> None:
        self.loader = loader
        super().__init__(loader)

    def __call__(self, yml: str) -> Yaml:
        """Parse a specific YAML string.

        Parameters
        ----------
        yml: str
            The YAML string to parse

        Returns
        -------
        dict or list
            The result of parsing the YAML string.

        """
        return yaml.load(yml, self.loader) or {}



"""Tools for loading, completing, and parsing text and configuration files.

Parameters that are known at program start are used to initialize the classes
so that, at runtime, the information only available then can flow through a
preconfigured pipe of callable objects to yield the desired result.

"""

from .resource import TextResourceLoader
from .interpolate import TemplateRenderer, FormFiller
from .read import TomlReader, YamlReader
from .parse import YamlParser
from .misc import NotFound

__all__ = [
    'TextResourceLoader',
    'TemplateRenderer',
    'FormFiller',
    'NotFound',
    'TomlReader',
    'YamlReader',
    'YamlParser'
]



from typing import Any, Self
from functools import cached_property
import torch as pt
import torch.nn as ptn
from ..types import Tensor, Module
from .activated import ActivatedEmbedder
from .gated import GatedEmbedder
from .gated_residual import GatedResidualEmbedder

type EmbCls = type[ActivatedEmbedder | GatedEmbedder | GatedResidualEmbedder]


class NumericalEmbedder(Module):
    """Transform (scalar) numerical features into embedding vectors.

    Parameters
    ----------
    mod_dim: int
        Desired embedding size. Will become the size of the last dimension of
        the output tensor.
    n_features: int
        Number of features to embed, which must equal the size of the last
        dimension of the input tensor.
    emb_cls: type
        The PyTorch module to use as embedding class. Must take `mod_dim` as
        its first argument on instantiation, take tensors of size 1 in their
        last dimension, and change that dimension to size `mod_dim`.
    **args
        Additional arguments to use when instantiating `emb_cls`.
    **kwargs
        Additional keyword arguments to use when instantiating `emb_cls`.

    See Also
    --------
    ActivatedEmbedder
    GatedEmbedder
    GatedResidualEmbedder

    """

    def __init__(
            self,
            mod_dim: int,
            n_features: int,
            emb_cls: EmbCls,
            *args: Any,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.n_features = n_features
        self.emb_cls = emb_cls
        self.args = args
        self.kwargs = kwargs
        self.embed = ptn.ModuleList(
            [emb_cls(mod_dim, *args, **kwargs) for _ in self.features]
        )

    @cached_property
    def features(self) -> range:
        """Range of feature indices."""
        return range(self.n_features)

    @cached_property
    def dim(self) -> int:
        """The output tensor dimension index to stack features into."""
        return -1 if self.n_features < 1 else -2

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass for embedding scalar numerical features into vectors.

        Parameters
        ----------
        inp: Tensor
            The last dimension of the input tensor is expected to be of size
            `n_features` and to contain the scalar values of the individual
            numerical features.

        Returns
        -------
        Tensor
            The output tensor has one more dimension of size `mod_dim` added
            after the last dimension (of size `n_features`) than the `inp`,
            containing the stacked embeddings.

        """
        emb = [self.embed[f](inp[..., [f]]) for f in self.features]
        return pt.stack(emb or self.mod_dim * [pt.zeros(*inp.shape)], self.dim)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        for emb in self.embed:
            emb.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            n_features: int | None = None,
            emb_cls: EmbCls | None = None,
            *args: Any,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Desired embedding size. Will become the size of the last dimension
            the output tensor. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        n_features: int, optional
            Number of features to embed, which must equal the size of the last
            dimension of the input tensor. Overwrites the `n_features` of the
            current instance if given. Defaults to ``None``.
        emb_cls: type, optional
            The PyTorch module to use as embedding class. Must take `mod_dim`
            as its first argument on instantiation. Overwrites the `emb_cls`
            of the current instance if given. Defaults to ``None``.
        *args
            Additional arguments replace those of the current instance and are
            then used when instantiating `emb_cls`.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then used together when
            instantiating `emb_cls`.

        Returns
        -------
        NumericalEmbedder
            A fresh, new instance of itself.

        See Also
        --------
        ActivatedEmbedder
        GatedEmbedder

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.n_features if n_features is None else n_features,
            self.emb_cls if emb_cls is None else emb_cls,
            *args,
            **(self.kwargs | kwargs)
        )



"""Pandas utilities and partials of dataframe method calls.

Parameters that are known at program start are used to initialize the classes
so that, at runtime, dataframes can flow through a preconfigured processing
pipe of callable objects.

"""

from .read import ParquetReader
from .frame import (
    AsType,
    ColumnSelector,
    ColumnsSelector,
    ColumnMapper,
    RowsSelector
)

__all__ = [
    'ParquetReader',
    'AsType',
    'ColumnSelector',
    'ColumnsSelector',
    'ColumnMapper',
    'RowsSelector'
]



"""Loggers specifically designed to keep your functions clean and tidy.

To avoid polluting your code, these loggers are meant to be inserted
into your functional workflow *between* its building blocks. Arguments passed
in are simply passed through to the next step, but you still have access to
them to customize the logged message.

"""

from .stdout import PassThroughStdOut, DEFAULT_FMT, PID_FMT

__all__ = [
    'PassThroughStdOut',
    'DEFAULT_FMT',
    'PID_FMT'
]



from typing import Any
from collections.abc import Callable, Sequence


class ReprName:
    """Auxiliary mixin providing representations for (callable) objects."""

    def _repr(self, obj: Any, _: int = 0) -> str:
        """Representation for any object."""
        return self._name(obj) if callable(obj) else repr(obj)

    @staticmethod
    def _name(obj: None | type | Callable) -> str:
        """Representation for callable objects used for, e.g., error messages.

        Parameters
        ----------
        obj: callable
            Any callable object, e.g., a class, function, method, or lambda

        Returns
        -------
        str
            A human-readable representation of the callable `obj`.

        """
        if obj is None:
            return 'None'
        if isinstance(obj, ReprName):
            return repr(obj)
        try:
            name = obj.__qualname__
        except AttributeError:
            name = obj.__class__.__name__ + '(...)'
        return 'lambda' if '<lambda>' in name else name


class IndentRepr(ReprName):
    """Base class for a representation with numbered and indented children.

    This class is not meant to be instantiated by itself. Rather, it is meant
    to be inherited from. Its constructor is then meant to be called in the
    child's constructor (with ``super().__init__(...)``), passing as arguments
    all items that are desired to appear in a (zero-based) numbered-list
    representation. If any of these objects have inherited from ``IndentRepr``
    themselves, their representation will recursively be indented by another
    level. Additional (keyword) arguments will be included in the first line
    of the representation.

    Parameters
    ----------
    items: sequence, optional
        Objects to represent in a (zero-based) numbered list below the class.
        Defaults to an emtpy tuple.
    *args
        Additional arguments to appear in the class instantiation signature
        in the first line of the representation.
    **kwargs
        Additional keyword arguments to appear in the class instantiation
        signature in the first line of the representation.

    Examples
    --------
    >>> class Fancy(IndentRepr):
    ...     def __init__(self, foo, *args, **kwargs) -> None:
    ...         super().__init__(args, foo, **kwargs)
    ...
    >>> Fancy('bar', 1, 2, Fancy('baz', 3, 4), 5, answer=42)
    Fancy('bar', answer=42):
    [ 0] 1
    [ 1] 2
    [ 2] Fancy('baz'):
         [ 0] 3
         [ 1] 4
    [ 3] 5

    """

    def __init__(
            self,
            items: Sequence = (),
            *args: Any,
            **kwargs: Any
    ) -> None:
        self.__items = items
        self.__args = args
        self.__kwargs = kwargs

    def __repr__(self) -> str:
        return self._indented_repr(0)

    def _repr(self, obj: Any, level: int = 0) -> str:
        """Return (potentially) indented object representations."""
        if isinstance(obj, IndentRepr):
            return obj._indented_repr(level + 1)
        return super()._repr(obj)

    def _indented_repr(self, level: int) -> str:
        """Construct indented object representation."""
        cls = self.__class__.__name__
        args = ', '.join(self._repr(arg) for arg in self.__args)
        kwargs = (f'{k}={self._repr(v)}' for k, v in self.__kwargs.items())
        kwargs = ', '.join(kwargs)
        signature = ', '.join(filter(None, [args, kwargs]))
        suffix = ":\n" if self.__items else ""
        indent = 5 * level * ' '
        items = enumerate(self._repr(item, level) for item in self.__items)
        items = '\n'.join(f'{indent}[{i:>2}] {item}' for i, item in items)
        return f'{cls}({signature})' + suffix + items


class ArgRepr(ReprName):
    """Base class for a representation with class name and (keyword) arguments.

    This class is not meant to be instantiated by itself. Rather, it is meant
    to be inherited from. Its constructor is then meant to be called in the
    child's constructor (with ``super().__init__(...)``), passing the (keyword)
    arguments that are desired to appear in the child's representation.

    Parameters
    ----------
    *args
        Appear as arguments within parentheses in the child representation
        after its class name.
    **kwargs
        Appear as keyword arguments within parentheses in the child
        representation after its class name and arguments.

    Examples
    --------
    >>> class Pretty(ArgRepr):
    ...     def __init__(self, foo, answer):
    ...         super().__init__(foo, answer=answer)
    ...
    >>> Pretty('good', 42)
    Pretty('good', answer=42)

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.__args = args
        self.__kwargs = kwargs

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        args = ', '.join(self._repr(arg) for arg in self.__args)
        kwargs = (f'{k}={self._repr(v)}' for k, v in self.__kwargs.items())
        kwargs = ', '.join(kwargs)
        signature = ', '.join(filter(None, [args, kwargs]))
        return f'{cls}({signature})'

    def _repr(self, obj: Any, _: int = 0) -> str:
        """Representation for any object."""
        if isinstance(obj, IndentRepr):
            head = repr(obj).splitlines()[0]
            head = head[:-1] if head.endswith(':') else head
            suffix = f'[{len(repr(obj).splitlines()) - 1}]'
            return head + suffix
        return super()._repr(obj)



"""Save and load (the state of) PyTorch models to and from disk."""

import warnings
import torch as pt
from pathlib import Path
from ..misc import ArgRepr
from ..text import NotFound
from .types import Module, Device

__all__ = [
    'StateSaver',
    'StateLoader',
    'ModelSaver',
    'ModelLoader'
]


class StateSaver(ArgRepr):
    """Save the state of a model to file.

    Parameters
    ----------
    path: str
        Path (including file name) to save a model's ``state_dict()`` to.
        May include any number of string placeholders (i.e., pairs of curly
        brackets) that will be interpolated when instances are called.

    """

    def __init__(self, path: str) -> None:
        self.path = str(Path(path.strip()).resolve())
        super().__init__(self.path)

    def __call__(self, model: Module, *parts: str) -> tuple[()]:
        """Save the state of a model, optimizer, or scheduler to file.

        Parameters
        ----------
        model: Module
            Model to save the state of.
        *parts: str, optional
            Fragments that will be interpolated into the `path` string given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `path`.

        Returns
        -------
        tuple
            An empty tuple.

        """
        file = self.path.format(*parts).strip()
        pt.save(model.state_dict(), file)
        return ()


class StateLoader(ArgRepr):
    """Load the state of a model from file.

    Parameters
    ----------
    path: str
        Path (including file name) to the  model's ``state_dict()`` on disk.
        May include any number of string placeholders (i.e., pairs of curly
        brackets) that will be interpolated when instances are called.
    map_location: str or Device, optional
        The device to load the state onto. Defaults to ``None`` which loads
        to the PyTorch device(s) that were saved with the model.
    merge: bool, optional
        Whether the loaded state should be merged into the state of the model
        (``True``) or replace its state (``False``). Defaults to ``True``.
    not_found: str, optional
        What to do if the specified file is not found. One of "ignore",
        "warn", or "raise" (use the ``NotFound`` enum to avoid typos).
        Defaults to "raise". If set to "ignore" or "warn" and the specified
        file is not found, `merge` is overridden to ``True``, thus returning
        the unaltered model.

    See Also
    --------
    NotFound

    """

    def __init__(
            self,
            path: str,
            map_location: Device | str | None = None,
            merge: bool = True,
            not_found: str = NotFound.RAISE
    ) -> None:
        self.path = str(Path(path.strip()).resolve())
        self.map_location = map_location
        self.merge = merge
        self.not_found = not_found.strip().lower()
        super().__init__(self.path, map_location, merge, self.not_found)

    def __call__(self, model: Module, *parts: str) -> Module:
        """Load the state of a model from file.

        Parameters
        ----------
        model: Module
            Model to load the state of.
        *parts: str, optional
            Fragments that will be interpolated into the `path` string given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `path`.

        Returns
        -------
        Module
            The `model` with its state restored.

        """
        file = self.path.format(*parts).strip()
        try:
            loaded = pt.load(file, self.map_location, weights_only=True)
            we_should_merge = self.merge
        except FileNotFoundError as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'File {} not found!\nReturning empty state dict.'
                    warnings.warn(msg.format(file))
                    loaded = {}
                    we_should_merge = True
                case NotFound.IGNORE:
                    loaded = {}
                    we_should_merge = True
                case _:
                    raise error
        state = model.state_dict() | loaded if we_should_merge else loaded
        _ = model.load_state_dict(state)
        return model.to(self.map_location) if hasattr(model, 'to') else model


class ModelSaver(ArgRepr):
    """Save an entire model to file.

    Parameters
    ----------
    path: str
        Path (including file name) to save the model to. May include any
        number of string placeholders (i.e., pairs of curly brackets) that will
        be interpolated when instances are called.

    """

    def __init__(self, path: str) -> None:
        self.path = str(Path(path.strip()).resolve())
        super().__init__(self.path)

    def __call__(self, model: Module, *parts: str) -> tuple[()]:
        """Save a model to file.

        Parameters
        ----------
        model: Module
            The model to save.
        *parts: str, optional
            Fragments that will be interpolated into the `path` string given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `path`.

        Returns
        -------
        tuple
            An empty tuple.

        """
        file = self.path.format(*parts).strip()
        pt.save(model, file)
        return ()


class ModelLoader(ArgRepr):
    """Load a previously saved model from file.

    Parameters
    ----------
    path: str, optional
        Full or partial path to the file to load. Will be interpreted as a
        directory and prepended as such if the `path` is completed when
        calling instances. Defaults to the current working directory of
        the python interpreter.
    map_location: str or device, optional
        The device to load the modelo onto. Defaults to ``None`` which loads
        to the PyTorch default device.

    """

    def __init__(
            self,
            path: str = '',
            map_location: Device | str | None = None,
    ) -> None:
        self.path = str(Path(path.strip()).resolve())
        self.map_location = map_location
        super().__init__(self.path, self.map_location)

    def __call__(self, path: str = '') -> Module:
        """Load a previously saved model from file.

        Parameters
        ----------
        path: str, optional
            Path (including file name) relative to the `path` specified at
            instantiation. Defaults to an empty string, which results in an
            unchanged `path` on concatenation.

        Returns
        -------
        Module
            The loaded model.

        """
        path = '/' + path.strip(' /') if path.strip(' /') else ''
        model = pt.load(self.path + path, self.map_location, weights_only=True)
        return model.to(self.map_location) if hasattr(model, 'to') else model



from typing import Any
from collections.abc import Callable


class FilterError(Exception):
    pass


class ForkError(Exception):
    pass


class MapError(Exception):
    pass


class PipeError(Exception):
    pass


class ReduceError(Exception):
    pass


class RouteError(Exception):
    pass


class SplitError(Exception):
    pass


class SumError(Exception):
    pass


class SideEffectsError(Exception):
    pass


class SafeError(Exception):
    """Special exception to wrap other exceptions.

    Parameters
    ----------
    error: Exception
        The wrapped exception.
    name: str
        A human-readable string representation of the callable that failed.
    call: callable
        The callable that failed.
    call_args: tuple
        The arguments that `call` failed with.

    See Also
    --------
    Safe

    """

    def __init__(
            self,
            error: Exception,
            name: str,
            call: type | Callable[..., Any],  # Not sure if we really need this
            call_args: tuple[Any, ...]
    ) -> None:
        self.error = error
        self.name = name
        self.call = call
        self.call_args = call_args
        super().__init__(self.message)

    @property
    def message(self):
        """The error message displayed if the exception is actually raised."""
        msg = '\nCaught {} calling\n{}\nwith arguments\n{}\n{}'
        err_cls = self.error.__class__.__name__
        return msg.format(err_cls, self.name, self.call_args, self.error)



from typing import Any, Self
from collections.abc import Iterator, Iterable
from functools import singledispatchmethod
import json
from json.decoder import JSONDecodeError
from ast import literal_eval
from pandas import Series, DataFrame
from .exceptions import ParseError, SchemaError, CastError, ValidationErrors
from .jsonobject import SchemaMeta, JsonObject

type Json = dict[str, Any]
type Record = str | bytes | bytearray | Json | Series | JsonObject | None
type Records = str | bytes | bytearray | DataFrame | Record | Iterable[Record]


class JsonObjects[T]:
    """List-like container for JSON-serializable dictionaries.

    This class is not meant to ever be instantiated directly. Rather,
    inherit from it, and specify a subclass of ``JsonObject`` with the
    `item_type` class keyword on sub-classing.

    Parameters
    ----------
    items : list, optional
        List of JSON-serializable dictionaries. Defaults to an empty list.
    *args
        Additionally, any number of JSON-serializable dictionaries with the
        schema specified by the `item_type` can be provided. These will be
        appended to the `items`.

    Raises
    ------
    ParseError
        If the constructor argument(s) can not be parsed as a list of
        dictionaries.
    SchemaError
        If `item_type` is not a subclass of ``JsonObject``.

    Warnings
    --------
    This class is rather heavy, so do not use it to, e.g., wrap JSON payloads
    in high-throughput low-latency web services!

    """

    def __init__(self, items: Records | Self = (), *args: Record) -> None:
        parsed = self.__parse(items)
        itemized = self.__itemize(parsed)
        self.__items = self.__wrap([*itemized, *args])

    def __init_subclass__(cls, **kwargs: Any) -> None:
        try:
            item_type = kwargs.pop('item_type')
        except KeyError:
            try:
                item_type = cls.mro()[1].__item_type__
            except AttributeError as error:
                msg = ('If not inherited, an "item_type" must be defined as'
                       'a keyword argument in the class call on definition!')
                raise SchemaError(msg) from error
        cls.__item_type__ = cls.__class_checked(item_type)
        super().__init_subclass__(**kwargs)

    def __str__(self) -> str:
        return json.dumps(self.__items, default=lambda obj: obj.as_json)

    def __repr__(self) -> str:
        return json.dumps(self.__items, indent=4, default=lambda o: o.as_json)

    def __iter__(self) -> Iterator[T]:
        return self.__items.__iter__()

    def __reversed__(self) -> Self:
        return self.__class__(reversed(self.__items))

    def __len__(self) -> int:
        return self.__items.__len__()

    def __getattr__(self, key: str) -> list:
        missing = object()
        values = [item.get(key, missing) for item in self.__items]
        if all(value is missing for value in values):
            cls = self.__class__.__name__
            msg = f"'{cls}' object has no attribute '{key}'"
            raise AttributeError(msg)
        return [None if value is missing else value for value in values]

    @singledispatchmethod
    def __getitem__(self, index: int) -> T:
        return self.__items[index]

    @__getitem__.register
    def _(self, index: slice) -> Self:
        return self.__class__(*self.__items[index])

    @__getitem__.register
    def _(self, key: str) -> list:
        missing = object()
        values = [item.get(key, missing) for item in self.__items]
        if all(value is missing for value in values):
            raise KeyError(key)
        return [None if value is missing else value for value in values]

    def __bool__(self) -> bool:
        return bool(self.__items)

    def __contains__(self, other: Record) -> bool:
        try:
            return self.__item_type__(other) in self.__items
        except (ParseError, CastError, ValidationErrors):
            return False

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.__items == other.__items
        return NotImplemented

    def __ne__(self: T, other: T) -> bool:
        if isinstance(other, self.__class__):
            return self.__items != other.__items
        return NotImplemented

    def __add__(self, others: Records | Self) -> Self:
        return self.__class__(self, *self.__class__(others))

    def __radd__(self, others: Records | Self) -> Self:
        return self.__class__(self.__class__(others), *self)

    def __call__(self, mapping: Record = None, **kwargs: Any) -> Self:
        """Update one or more (nested) fields within each item.

        Parameters
        ----------
        mapping: dict or str, optional
            Dictionary with string keys, JSON string/bytes, or pandas Series.
            Defaults to an empty dictionary.
        **kwargs:
            Can be any value or, for nested structures, again a dictionary with
            string keys or a JSON string/bytes or a pandas Series. Keyword
            arguments will override values already present in the `mapping`.

        Returns
        -------
        JsonObjects
            A new instance of self with updated fields in each item.

        Raises
        ------
        ParseError
            If the (keyword) arguments cannot be parsed into a dictionary with
            string keys.
        CastError
            If the dictionary values cannot be cast into the types specified in
            the schema of the `item_type`.

        """
        return self.__class__(item(mapping, **kwargs) for item in self)

    @property
    def as_json(self) -> list[Json]:
        """JSON-serializable representation."""
        return [item.as_json for item in self.__items]

    @property
    def as_dtype(self) -> str:
        """Representation in the cell of a pandas data frame."""
        return self.__str__()

    @property
    def as_df(self) -> DataFrame:
        """Representation as a pandas data frame."""
        data = [item.as_series for item in self]
        if data:
            columns = None
        else:
            columns = list(self.__item_type__.__annotations__.keys())
        df = DataFrame(data, columns=columns)
        df.columns.name = self.__item_type__.__name__
        return df.reset_index(drop=True)

    @staticmethod
    def __class_checked(item_type: type[JsonObject]) -> type[JsonObject]:
        """Allow only JsonObject and JsonObjects as item_type."""
        right_type = isinstance(item_type, SchemaMeta)
        right_class = issubclass(item_type, JsonObject)
        if right_type and right_class:
            return item_type
        raise SchemaError('item_type must be a subclass of JsonObject!')

    @staticmethod
    def __parse(items: Records) -> list[Record]:
        """Parse input into a list of something."""
        # Define parsers for converting input into a list of items.
        parsers = (
            lambda x: json.loads(x),                # JSON string
            lambda x: literal_eval(x),              # Some other string
            lambda x: x.to_dict(orient='records'),  # Dataframe
            lambda x: [] if x is None else x        # None or some other object
        )
        # Try parsers one after another
        for parse in parsers:
            try:
                parsed = parse(items)
            except (
                JSONDecodeError,  # json.loads
                TypeError,        # json.loads and literal_eval
                ValueError,       # literal_eval
                SyntaxError,      # literal_eval
                AttributeError    # Dataframe
            ):
                continue
            else:
                return parsed
        raise ParseError(f'Could not parse {items} as JSON!')

    @staticmethod
    def __itemize(items: list[Record]) -> list[Json]:
        """Convert list of something into a list of JSONs."""
        # Define patterns that convert input to a list of dicts
        patterns = (
            lambda x: [{**x}],  # List of dicts
            lambda x: [*x]      # List of iterable, possibly a string
        )
        # First, try list of dicts, then list of anything
        for pattern in patterns:
            try:
                itemized = pattern(items)
            except TypeError:
                continue
            else:
                return itemized
        raise ParseError(f'Could not parse {items} as JSON!')

    def __wrap(self, items: list[Record]) -> list[T]:
        """Cast each item in a list of JSONs to the item schema."""
        errors = []
        wrapped = []
        for item in items:
            try:
                wrapped.append(self.__item_type__(item))
            except (ParseError, CastError, ValidationErrors) as error:
                errors.append(error)
        if errors:
            raise ValidationErrors(self.__class__.__name__, errors)
        return wrapped



"""Flexible data structures, ideal to specify nested project configurations."""

from .jsonobject import JsonObject
from .jsonobjects import JsonObjects

__all__ = [
    'JsonObject',
    'JsonObjects'
]



"""Flexibly parameterized and feature-rich training loop for models.

Allows for warm-up periods as well as early stopping with recovery of the best
checkpoint from CPU memory or disk. Use the provided callback examples to
monitor training progress or derive your own from the respective base classes.
Custom templates for train and test data to seamlessly work with the training
loop are also provided.

"""

from .data import TestDataBase, TrainDataBase
from .trainer import Trainer
from .checkpoints import Checkpoint, InMemory, OnDisk, State
from .callbacks import (
    EpochPrinter,
    TrainPrinter,
    EpochCallback,
    History
)

__all__ = [
    'Trainer',
    'InMemory',
    'OnDisk',
    'EpochPrinter',
    'TrainPrinter',
    'TrainDataBase',
    'TestDataBase',
    'Checkpoint',
    'EpochCallback',
    'State',
    'History'
]



from importlib import import_module
from ..misc import ArgRepr
from .exceptions import ImporterError


class Importer(ArgRepr):
    """Programmatically import objects from a module under a top-level package.

    For ease of use and clarity in API, relative imports are not supported.
    Objects are instantiated with where to import from and called with what
    to import.

    Parameters
    ----------
    package: str
        Name of the top-level package to import from. Must not start with dots
        but can contain any number of dots to indicate sub-packages.
    module: str, optional
        The specific module to import objects from. May contain dots to
        indicate that it is located further down within some sub-package.
        Defaults to "steps".

    """

    def __init__(self, package: str, module: str = 'steps') -> None:
        self.package = package.strip(' ./')
        self.module = module.strip(' ./')
        super().__init__(self.package, self.module)

    @property
    def path(self) -> str:
        """Full path specification of (sub-)package and module concatenated."""
        return '.'.join([self.package, self.module])

    def __call__(self, *names: str) -> list:
        """Import any number of objects from the specified package.module.

        Parameters
        ----------
        *names: str
            Name(s) of object(s) to import from ``package.module``.

        Returns
        -------
        list
            Imported objects.

        Raises
        ------
        ImporterError
            When the ``package.module`` is mis-specified, can't be found, or
            when the specified object(s) can't be found in it.

        """
        try:
            location = import_module(self.path)
        except (TypeError, ModuleNotFoundError) as error:
            msg = 'Could not import module "{}"!'
            raise ImporterError(msg.format(self.path)) from error
        imports = []
        for name in names:
            try:
                imported = getattr(location, name)
            except AttributeError as err:
                msg = 'Could not import "{}" from module "{}"!'
                raise ImporterError(msg.format(name, self.path)) from err
            imports.append(imported)
        return imports



import time
import uuid
import os
from typing import Any
from google.cloud import bigquery as gbq
from google.cloud import storage as gcs
from .exceptions import GbqError
from ...misc import ArgRepr


class GbqQuery2GcsParquet(ArgRepr):
    """Export the results of an SQL query to a bucket on Google Cloud Storage.

    The SQL query will be fired against Google BigQuery. In essence, an
    ``EXPORT`` clause with the specified parameters interpolated is inserted
    into the query, after the last semicolon, before everything else.
    As such, no semicolon-separated sub-queries are allowed, but variable
    declaration and setting is fine. Results will be saved as multiple,
    sequentially numbered, `snappy`-compressed parquet files.

    Parameters
    ----------
    project: str
        The name of the Google BigQuery billing project.
    bucket: str
        The Google Cloud Storage bucket to export to. Note that, unless you
        have set up some different "wiring" yourself, the project of the
        bucket is the same as `project`.
    prefix: str, optional
        Prefix of the blob location, where parquet files will reside. If
        none is given here, one can be provided on calling the instance.
        If both are given, they will be concatenated and, if neither
        is given, a UUID will be generated. Defaults to an empty string.
    overwrite: bool, optional
        Blobs with the given bucket/prefix combination may already exist on
        Google Cloud Storage. If ``True`` these are overwritten, else an
        exception is raised. Defaults to ``False``.
    skip: bool, optional
        Blobs with the given bucket/prefix combination may already exist on
        Google Cloud Storage. If that is the case, and `skip` is ``True``,
        nothing will be done at all. Defaults to ``False``
    polling_interval: int, optional
        Job completion is going to be checked for every `polling_interval`
        seconds. Defaults to 5 (seconds).
    priority: str, optional
        Priority the query job should be run as. Can be either "BATCH" or
        "INTERACTIVE". Defaults to "BATCH". Use the `QueryPriority
        <https://cloud.google.com/python/docs/reference/bigquery/latest/
        google.cloud.bigquery.job.QueryPriority>`__ enum to avoid typos.
    gbq_kws: dict, optional
        Additional keyword arguments to passed to the Google BigQuery client.
        Defaults to ``None``, which results in an empty dictionary.
    gcs_kws: dict, optional
        Additional keyword arguments to passed to the Google Storage client.
        Defaults to ``None``, which results in an empty dictionary.

    """

    _template = """    EXPORT DATA OPTIONS(
        uri="gs://{}/{}*.parquet"
      , format="PARQUET"
      , compression="SNAPPY"
      , overwrite={}
    ) AS"""

    def __init__(
            self,
            project: str,
            bucket: str,
            prefix: str = '',
            overwrite: bool = False,
            skip: bool = False,
            polling_interval: int = 5,
            priority: str = 'BATCH',
            gbq_kws: dict[str, Any] | None = None,
            gcs_kws: dict[str, Any] | None = None,
    ) -> None:
        self.project = project.strip(' /.')
        self.bucket = bucket.strip(' /.')
        self.prefix = prefix.strip(' /.')
        self.overwrite = overwrite
        self.skip = skip
        self.polling_interval = polling_interval
        self.priority = priority.strip().upper()
        self.gbq_kws = {} if gbq_kws is None else gbq_kws
        self.gcs_kws = {} if gcs_kws is None else gcs_kws
        super().__init__(
            self.project,
            self.bucket,
            self.prefix,
            self.overwrite,
            self.skip,
            self.polling_interval,
            self.priority,
            self.gbq_kws,
            self.gcs_kws,
        )

    def __call__(self, query: str, prefix: str = '', **kwargs: Any) -> str:
        """Export the results of a SQL query to Google Cloud Storage.

        Parameters
        ----------
        query: str
            The SQL query to fire using the pre-configured client.
        prefix: str, optional
            Prefix of the blob location, where parquet files will reside. If
            none is given, the prefix specified at instantiation will be used.
            If both are given, they will be concatenated and, if neither
            is given, a UUID will be generated. Defaults to an empty string.
        **kwargs
            Additional keyword arguments are passed to the constructor of the
            Google BigQuery ``QueryJobConfig``. See `documentation
            <https://cloud.google.com/python/docs/reference/bigquery/latest/
            google.cloud.bigquery.job.QueryJobConfig>`__ for options.

        Returns
        -------
        str
            If the query finishes without errors, the given or generated prefix
            is returned, so that blobs with the exported data can be retrieved.

        Raises
        ------
        GbqError
            If the submitted ``QueryJob`` finishes and returns and error.

        """
        scripts, main = self.__split(query)
        prefix, header = self.__render(prefix.strip(' ./'))
        if self.__skip_query_for(prefix):
            return prefix
        client = gbq.Client(self.project, **self.gbq_kws)
        config = gbq.QueryJobConfig(priority=self.priority, **kwargs)
        job = client.query('\n'.join([scripts, header, main]), config)
        while job.running():
            time.sleep(self.polling_interval)
        if error := job.error_result:
            raise GbqError(f"\n{error['reason'].upper()}: {error['message']}")
        return prefix

    def __skip_query_for(self, prefix: str) -> bool:
        """Do blobs with prefix exist on Google Storage? Overwrite them?"""
        client = gcs.Client(self.project, **self.gcs_kws)
        blobs = list(client.list_blobs(self.bucket, prefix=prefix))
        if blobs and self.skip:
            return True
        if blobs and self.overwrite:
            bucket = client.get_bucket(self.bucket)
            bucket.delete_blobs(blobs)
        return False

    def __render(self, prefix: str) -> tuple[str, str]:
        """Render the EXPORT header template with the given parameters."""
        prefix = os.path.join(self.prefix, prefix) if prefix else self.prefix  # noqa: PTH118
        prefix = prefix or str(uuid.uuid4())
        prefix = prefix + '/' if prefix else ''
        header = self._template.format(
            self.bucket,
            prefix,
            str(self.overwrite).lower()
        )
        return prefix, header

    @staticmethod
    def __split(query: str) -> tuple[str, str]:
        """Split query into scripts (before last semicolon) and main part."""
        split = query.split(';')
        separator = ';\n' if len(split) > 1 else ''
        scripts = ';'.join(split[:-1]) + separator
        return scripts, split[-1]



import sys
import json
from json import JSONDecodeError
from ast import literal_eval
from argparse import ArgumentParser, RawTextHelpFormatter
from typing import Any
from itertools import takewhile, dropwhile, chain
from .exceptions import ArgParseError

type Parsed = tuple[list[str], dict[str, Any]]

USAGE = '%(prog)s [action(s)] [-h]'

DESCRIPTION = 'Refer to the README.md for the available actions!'

EPILOG = """
Additionally, all fields of the program's config can be set via
long-format options. Nested fields can be set by dot-separating
levels, e.g., "--root.level1.level2 value"

{!r}
"""


class ArgParser:
    """Parse the command line for actions and any long-format options.

    Using this command-line argument parser alleviates the need for
    defining any groups or options beforehand. Arguments immediately
    following the program call are interpreted as actions to perform as long
    as they do not start with a hyphen. Starting with the first argument
    that starts with a hyphen, command-line arguments will be interpreted as
    ``--key value`` pairs and this long format is the only one allowed.
    Abbreviated options (``-k value``) right after actions (and before any
    long-format options) are ignored.

    Parameters
    ----------
    default_action: str, optional
        Default action to return if none is found in the command-line
        arguments. Defaults to no action.
    usage: str, optional
        Program usage message.
    description: str, optional
        Program description.
    epilog: str, optional
        Text displayed after the help on command-line options.
    fmt_cls: type, optional
        Option passed on to the underlying ``argparse.ArgumentParser``.
        Defaults to ``argparse.RawTextHelpFormatter``

    """

    def __init__(
            self,
            default_action: str | None = None,
            usage: str = USAGE,
            description: str = DESCRIPTION,
            epilog: str = EPILOG,
            fmt_cls: type = RawTextHelpFormatter
    ) -> None:
        self.default_action = default_action
        self.usage = usage
        self.description = description
        self.epilog = epilog
        self.fmt_cls = fmt_cls
        self.__parse = ArgumentParser(
            usage=usage,
            description=description,
            epilog=epilog,
            formatter_class=fmt_cls
        ).parse_known_args

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        return f'{cls}({self.default_action}, ...)'

    def __call__(self, args: list[str] | None = None) -> Parsed:
        """Parse the command-line arguments into actions and options.

        Parameters
        ----------
        args: list of str, optional
            The command-line arguments to parse. Mainly a debugging feature.
            If none is given, ``sys.argv[1:]`` will be parsed.

        Returns
        -------
        actions: list of str
            A list with the actions (as strings) to perform. If none are
            found on the command line and no `default_action` is specified,
            that list will be empty.
        options: dict
            Dictionary with keys and values parsed from long-format
            command line arguments.

        """
        args = sys.argv[1:] if args is None else args
        _, unknowns = self.__parse(args)
        actions, args = self.__split(unknowns)
        actions = self.__valid(actions)
        args = self.__compatible(args)
        options = self.__mapped(args)
        return actions, options

    def __split(self, args: list[str]) -> tuple[list[str], list[str]]:
        """Split command-line options into actions and config settings."""
        # Everything before the first option (starting with "-") are actions.
        actions = tuple(takewhile(lambda arg: not arg.startswith('-'), args))
        # If there are none, fall back onto the default action (if specified).
        if not actions and self.default_action is not None:
            actions = self.default_action,
        # Either way, replace dashes with underscores
        actions = (action.lower().replace('-', '_') for action in actions)

        # Everything after the action(s) are options (starting with "--").
        args = dropwhile(lambda arg: not arg.startswith('--'), args)
        # In case options are given as "--key=value", we split.
        args = chain.from_iterable(arg.split('=') for arg in args)
        return list(actions), list(args)

    @staticmethod
    def __valid(actions: list[str]) -> list[str]:
        """Raise if any action string is not a valid python identifier."""
        for action in actions:
            if not action.isidentifier():
                msg = 'Actions must be valid identifiers, unlike "{}"!'
                raise ArgParseError(msg.format(action))
        return actions

    @staticmethod
    def __compatible(args: list[str]) -> list[str]:
        """Raise if a command-line option is not in long format with value."""
        long_form = all(arg.startswith('--') for arg in args[::2])
        alternating = all(not arg.startswith('-') for arg in args[1::2])
        even_number = len(args) % 2 == 0
        if not (long_form and alternating and even_number):
            msg = ('Command-line arguments must be passed in long format (i.e.'
                   ', as "--key value"), and a value must always be present!')
            raise ArgParseError(msg)
        return args

    def __mapped(self, args: list[str]) -> dict[str, Any]:
        """Create dictionary wih config keys and (string) values from args."""
        zipped = list(zip(args[::2], args[1::2]))
        # Drop the first two characters (always "--") and replace "-" with "_".
        return {k[2:].replace('-', '_'): self.__parsed(v) for k, v in zipped}

    @staticmethod
    def __parsed(value: str) -> Any:
        """Try to parse (string) command-line options into python objects."""
        try:
            parsed = json.loads(value)
        except (TypeError, JSONDecodeError):
            try:
                parsed = literal_eval(value)
            except (TypeError, ValueError, SyntaxError):
                parsed = value
        return parsed



from typing import Any
import pandas_gbq as pq
from pandas import DataFrame
from ...misc import ArgRepr


class GbqQuery2DataFrame(ArgRepr):
    """Partial of the ``read_gbq`` function in the ``pandas_gbq`` package.

    As such, it may not bet suitable for downloading large amounts of data (see
    its `documentation <https://googleapis.dev/python/pandas-gbq/latest/>`__
    for details).

    Parameters
    ----------
    project: str
        The project to bill for the query/retrieval.
    **kwargs
        Additional keyword arguments passed on to the ``read_gbq`` call.

    See Also
    --------
    GbqQuery2GcsParquet
    GcsParquet2DataFrame
    GcsDir2LocalDir

    """

    def __init__(self, project: str, **kwargs: Any) -> None:
        self.project = project.strip(' ./')
        self.kwargs = kwargs
        super().__init__(self.project, **kwargs)

    def __call__(self, query_or_table: str) -> DataFrame:
        """Read Google BigQuery SQL results or table into a pandas DataFrame.

        Parameters
        ----------
        query_or_table: str
            Table name (including dataset id) or SQL query to be
            retrieved from or submitted to Google BigQuery.

        Returns
        -------
        DataFrame
            The contents of the table or the results of the SQL query.

        """
        return pq.read_gbq(
            query_or_table,
            project_id=self.project,
            progress_bar_type=None,
            **self.kwargs
        )



"""Create tensors from other array-like data, and change device and type."""

from typing import Any, overload
import torch as pt
from ..misc import ArgRepr
from .types import Tensor, Module, Dtype, Device, DataFrame

__all__ = [
    'Create',
    'AsTensor',
    'from_dataframe',
    'To'
]


class Create(ArgRepr):
    """Partial of the top-level PyTorch function ``tensor``.

    Parameters
    ----------
    dtype: dtype, optional
        Torch dtype of the tensor to create. Defaults to ``None``
    device: device, optional
        Torch device to create the tensor on. Defaults to ``None``
    requires_grad: bool, optional
        If autograd should record operations on the returned tensor.
        Defaults to ``False``.
    pin_memory: bool, optional
        If set, returned tensor would be allocated in the pinned memory.
        Works only for CPU tensors. Defaults to ``False``.

    """

    def __init__(
            self,
            dtype: Dtype | None = None,
            device: Device | str | None = None,
            requires_grad: bool = False,
            pin_memory: bool = False
    ) -> None:
        super().__init__(dtype, device, requires_grad, pin_memory)
        self.dtype = dtype
        self.device = device
        self.requires_grad = requires_grad
        self.pin_memory = pin_memory

    def __call__(self, data: Any) -> Tensor:
        """Create a PyTorch tensor with the cached (keyword) arguments.

        Parameters
        ----------
        data
            Any type of data a PyTorch tensor can be created from.

        Returns
        -------
        Tensor
            A PyTorch tensor.

        """
        return pt.tensor(
            data,
            dtype=self.dtype,
            device=self.device,
            requires_grad=self.requires_grad,
            pin_memory=self.pin_memory
        )


class AsTensor(ArgRepr):
    """Partial of the top-level PyTorch function ``as_tensor``.

    Parameters
    ----------
    dtype: dtype, optional
        Torch dtype of the tensor to create. Defaults to ``None``
    device: device, optional
        Torch device to create the tensor on. Defaults to ``None``

    """

    def __init__(
            self,
            dtype: Dtype | None = None,
            device: Device | str | None = None
    ) -> None:
        super().__init__(dtype, device)
        self.dtype = dtype
        self.device = device

    def __call__(self, data: Any) -> Tensor:
        """Create a PyTorch tensor with the cached (keyword) arguments.

        Parameters
        ----------
        data
            Any type of data a PyTorch tensor can be created from.

        Returns
        -------
        Tensor
            A PyTorch tensor sharing the memory with ``data``.

        """
        return pt.as_tensor(data, dtype=self.dtype, device=self.device)


def from_dataframe(df: DataFrame) -> Tensor:
    """Convert a pandas dataframe to a PyTorch tensor.

    This is simply wrapper around the top-level PyTorch function ``from_numpy``
    function called on the dataframe's ``values`` attribute.

    Parameters
    ----------
    df: DataFrame
        The pandas dataframe to convert.

    Returns
    -------
    Tensor
        A PyTorch tensor sharing the memory with the dataframe's data.

    """
    return pt.from_numpy(df.values)



class To(ArgRepr):
    """Move or change a tensor or module to a different device or dtype.

    Refer to the PyTorch `documentation <https://pytorch.org/docs/stable/
    generated/torch.Tensor.to.html#torch.Tensor.to>`__ for more information.

    Parameters
    ----------
    target: device or dtype
        The device or dtype to move the tensor or module to.
    *args
        Additional argument to pass to the `to` method.
    **kwargs
        Additional keyword arguments to pass to the `to` method.

    """

    def __init__(
            self,
            target: str | Device | Dtype,
            *args: Any,
            **kwargs: Any
    ) -> None:
        super().__init__(target, *args, **kwargs)
        self.target = target
        self.args = args
        self.kwargs = kwargs

    @overload
    def __call__(self, inp: Tensor) -> Tensor:
        ...

    @overload
    def __call__(self, inp: Module) -> Module:
        ...

    def __call__(self, inp):
        """Move or change a tensor or module to the specified target.

        Parameters
        ----------
        inp: Tensor or Module
            The input tensor or module to move or change to the `target`.

        Returns
        -------
        Tensor or Module
            The `inp` moved or changed to the cached `target`.

        """
        return inp.to(self.target, *self.args, **self.kwargs)



from abc import ABC, abstractmethod
from typing import Any, TypedDict
from collections import OrderedDict
from pathlib import Path
import torch as pt
from ..types import Tensor, Module, Optimizer, LRScheduler

__all__ = [
    'State',
    'Checkpoint',
    'InMemory',
    'OnDisk'
]

class State(TypedDict):
    """Container for a snapshot of the state of model training."""
    epoch: int  #: The current epoch.
    loss: float  #: The (train or test) loss at the current epoch.
    model: OrderedDict[str, Any]  #: The state-dict of the model.
    optimizer: dict[str, Any]  #: The state-dict of the optimizer.
    scheduler: dict[str, Any]  #: The state-dict of the scheduler.


class Checkpoint(ABC):
    """Base class to inherit from when implementing custom checkpoints."""

    def __init__(self) -> None:
        self.__counter = 0

    @property
    def counter(self) -> int:
        """How many checkpoints were saved since the last reset."""
        return self.__counter

    def save(
            self,
            epoch: int,
            loss: float,
            model: Module,
            optimizer: Optimizer | None = None,
            scheduler: LRScheduler | None = None
    ) -> None:
        """User-facing method for saving a checkpoint during training.

        Parameters
        ----------
        epoch: int
            The current epoch.
        loss: float
            The (train or test) loss at the current `epoch`.
        model: Module
            The model to checkpoint the state of.
        optimizer: Optimizer, optional
            The optimizer to checkpoint the state of. Defaults to ``None``.
        scheduler: LRScheduler, optional
            The scheduler to checkpoint the state of. Defaults to ``None``.

        """
        self.__counter += 1
        self._save_state({
            'epoch': epoch,
            'loss': loss,
            'model': model.state_dict(),
            'optimizer': {} if optimizer is None else optimizer.state_dict(),
            'scheduler': {} if scheduler is None else scheduler.state_dict()
        })

    def load(
            self,
            model: Module,
            optimizer: Optimizer | None = None,
            scheduler: LRScheduler | None = None
    ) -> tuple[int, float]:
        """User-facing method for loading a checkpoint during training.

        The state retrieved from the checkpoint is merged into the state of
        the objects to update, such that loading a checkpoint before one has
        been saved results in unchanged objects.

        Parameters
        ----------
        model: Module
            The model to load the state for in-place.
        optimizer: Optimizer, optional
            The optimizer to load the state for in-place. Defaults to ``None``.
        scheduler: LRScheduler, optional
            The scheduler to load the state for in-place. Defaults to ``None``.

        Returns
        -------
        epoch: int
            The epoch of the checkpoint.
        loss: float
            The loss from the checkpoint.

        """
        buffer = self._load_state()
        model.load_state_dict(model.state_dict() | buffer['model'])
        if optimizer is not None:
            optimizer.load_state_dict(
                optimizer.state_dict() | buffer['optimizer']
            )
        if scheduler is not None:
            scheduler.load_state_dict(
                scheduler.state_dict() | buffer['scheduler']
            )
        return buffer['epoch'], buffer['loss']

    def reset_parameters(self) -> None:
        """Hard-resets the checkpoint into a pristine, initial state."""
        self.__counter = 0
        self._save_state({
            'epoch': 0,
            'loss': float('inf'),
            'model': OrderedDict(),
            'optimizer': {},
            'scheduler': {}
        })

    @abstractmethod
    def _save_state(self, state: State) -> None:
        """Subclasses must implement how exactly the state is persisted.

        Parameters
        ----------
        state: State
            The state to persist.

        Warnings
        --------
        Make sure the persisted state is decoupled from the actual
        state (e.g., copied) so that it does not change as model
        training continues!

        """
        ...

    @abstractmethod
    def _load_state(self) -> State:
        """Subclasses must implement how exactly the state is retrieved.

        Returns
        -------
        State

        Warnings
        --------
        Make sure that the returned state is decoupled from the persisted
        state (e.g., copied) so that the latter cannot change as model
        training continues!

        """
        ...


class InMemory(Checkpoint):
    """Checkpoint in CPU memory regardless of the device training runs on."""

    def __init__(self) -> None:
        super().__init__()
        self.state = {
            'epoch': 0,
            'loss': float('inf'),
            'model': OrderedDict(),
            'optimizer': {},
            'scheduler': {}
        }

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'

    def _to_cpu(self, obj: Any) -> Any:
        """Recursively copy tensors to CPU and remember original devices."""
        if isinstance(obj, dict):
            clone = {}
            for key, value in obj.items():
                clone[key] = self._to_cpu(value)
            return clone
        if isinstance(obj, list):
            return [self._to_cpu(item) for item in obj]
        if isinstance(obj, Tensor):
            clone = pt.empty_like(obj, device='cpu').copy_(obj, True)
            clone.target = obj.device
            return clone
        return obj

    def _to_device(self, obj: Any) -> Any:
        """Recursively copy backed-up tensors to their original devices."""
        if isinstance(obj, dict):
            clone = {}
            for key, value in obj.items():
                clone[key] = self._to_device(value)
            return clone
        if isinstance(obj, list):
            return [self._to_device(item) for item in obj]
        if isinstance(obj, pt.Tensor):
            return pt.empty_like(obj, device=obj.target).copy_(obj, True)
        return obj

    def _save_state(self, state: State) -> None:
        """Create a deep copy of the combined state in CPU memory."""
        self.state = self._to_cpu(state)

    def _load_state(self) -> State:
        """Create a deep copy of the saved state on the original device(s)."""
        return self._to_device(self.state)


class OnDisk(Checkpoint):
    """Checkpoint the current state of training on disk.

    Parameters
    ----------
    path: str
        Full path to and name of the file used to persist state.

    """

    def __init__(self, path: str) -> None:
        super().__init__()
        self.path = str(Path(path.strip()).resolve())
        self.reset_parameters()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.path}')"

    def _save_state(self, state: State) -> None:
        """Save the combined state to file."""
        pt.save(state, self.path)

    def _load_state(self) -> State:
        """Retrieve the combined state from file."""
        return pt.load(self.path, weights_only=True)



"""PyTorch probability distributions re-parameterized for convenience.

The output of networks designed to provide probabilistic predictions becomes
much more interpretable when they predict the mean and the standard deviation
of some probability distribution rather than its (often arcance) natural
parameters. To that end, this module provides re-parameterized versions
of probability distributions to match the parameterization of the respective
log-likelihoods implemented as loss functions.

"""

import torch.distributions as ptd
from .types import Tensor

__all__ = [
    'MuSigmaGamma',
    'MuSigmaNegativeBinomial'
]


class MuSigmaGamma(ptd.Gamma):
    """Subclass of the PyTorch `Gamma` distribution with intuitive parameters.

    Parameters
    ----------
    mu: Tensor
        Mean value(s) of the Gamma distribution(s).
    sigma: Tensor
        Standard deviation(s) of the Gamma distribution(s).
    validate_args: bool, optional
        Whether the parent class should validate the transformed parameters
        or not. Defaults to ``False``

    See Also
    --------
    swak.pt.losses.GammaLoss

    """

    def __init__(
            self,
            mu: float | Tensor,
            sigma: float | Tensor,
            validate_args: bool = False
    ) -> None:
        variance = sigma**2
        super().__init__(mu**2 / variance, mu / variance, validate_args)


class MuSigmaNegativeBinomial(ptd.NegativeBinomial):
    """Subclass of the PyTorch `NegativeBinomial` distribution.

    Parameters
    ----------
    mu: Tensor
        Mean value(s) of the Gamma distribution(s).
    sigma: Tensor
        Standard deviation(s) of the Gamma distribution(s).
    validate_args: bool, optional
        Whether the parent class should validate the transformed parameters
        or not. Defaults to ``False``

    Notes
    -----
    This parameterization only makes sense if the variance is strictly
    greater than the mean. This is best taken into account already on the
    model output side, but can be checked here by setting `validate_args`
    to ``True``.

    See Also
    --------
    swak.pt.losses.NegativeBinomialLoss
    swak.pt.misc.NegativeBinomialFinalizer

    """

    def __init__(
            self,
            mu: float | Tensor,
            sigma: float | Tensor,
            validate_args: bool = False
    ) -> None:
        sigma2 = sigma**2
        super().__init__(
            mu**2 / (sigma2 - mu),
            1.0 - mu / sigma2,
            validate_args=validate_args
        )



from typing import Any, Self
import torch.nn as ptn
from ..types import Tensor, Module, Functional, Drop


class GatedResidualConcatMixer(Module):
    """Combined stacked feature vectors through a Gated Residual Network (GRN).

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension and the output will again have this size in
        its last dimension.
    n_features: int
        The number of features to combine. Must be equal to the size of the
        next-to-last dimension of the input tensor.
    activate: Module or function, optional
        The activation function to be applied after (linear) transformation,
        but prior to gating. Must be a callable that accepts a tensor as sole
        argument, like a module from ``torch.nn`` or a function from
        `torch.nn.functional``, depending on whether it needs to be further
        parameterized or not. Defaults to an ``ELU`` activation.
    gate: Module or function, optional
        The activation function to be applied to half of the (non-linearly)
        transformed input before multiplying with the other half. Must be
        a callable that accepts a tensor as sole argument, like a module from
        ``torch.nn`` or a function from ``torch.nn.functional``, depending
        on whether it needs to be further parameterized or not.
        Defaults to a sigmoid.
    drop: Module, optional
        Typically an instance of ``Dropout`` or ``AlphaDropout``. Defaults to
        ``Dropout(p=0.0)``, resulting in no dropout being applied.
    **kwargs
        Additional keyword arguments to pass through to the linear layers.

    Notes
    -----
    This implementation is inspired by how features are encoded in `Temporal
    Fusion Transformers`, [1]_ but it is not quite the same. Specifically, the
    intermediate linear layer (Eq. 3) is eliminated and dropout is applied
    directly to the activations after the first layer. Also, the layer norm
    (Eq. 2) is replaced by simply dividing the sum of (linearly projected)
    input and gated signal by 2. Should additional normalization be desired,
    it can be performed independently on the output of this module.

    References
    ----------
    .. [1] B. Lim, S. O. Arik, N. Loeff, and T. Pfister, `Temporal Fusion
           Transformers for Interpretable Multi-horizon Time Series
           Forecasting`, `arXiv:1912.09363v3 <https://arxiv.org/abs/
           1912.09363>`__ (2020).

    """

    def __init__(
            self,
            mod_dim: int,
            n_features: int,
            activate: Module | Functional = ptn.ELU(),
            gate: Module | Functional = ptn.Sigmoid(),
            drop: Drop = ptn.Dropout(0.0),
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.n_features = n_features
        self.activate = activate
        self.gate = gate
        self.drop = drop
        self.kwargs = kwargs
        self.project = ptn.Linear(n_features * mod_dim, mod_dim, **kwargs)
        self.widen = ptn.Linear(mod_dim, 2 * mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass for combining multiple stacked feature vectors.

        Parameters
        ----------
        inp: Tensor
            The size of the next-to-last last dimension of the input tensor is
            expected to match the `n_features` provided at instantiation.
            The last dimension (of size `mod_dim`) is expected to contain the
            features vectors themselves.

        Returns
        -------
        Tensor
            The output tensor has one fewer dimensions than the input.
            The next-to-last dimension is dropped and the size of the last
            dimension is once again `mod_dim`.

        """
        projected = self.project(inp.flatten(start_dim=-2))
        wide = self.widen(self.drop(self.activate(projected)))
        gated = wide[..., :self.mod_dim] * self.gate(wide[..., self.mod_dim:])
        return 0.5 * (projected + gated)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.project.reset_parameters()
        self.widen.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            n_features: int | None = None,
            activate: Module | Functional | None = None,
            gate: Module | Functional | None = None,
            drop: Drop | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Size of the feature space. The input tensor is expected to be of
            that size in its last dimension and the output will again have this
            size in its last dimension. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        n_features: int, optional
            The number of features to combine. Must be equal to the size of the
            next-to-last dimension of the input tensor. Overwrites `n_features`
            of the current instance if given. Defaults to ``None``.
        activate: Module or function, optional
            The activation function to be applied after (linear) transform,
            but prior to gating. Must be a callable that accepts a tensor as
            sole argument, like a module from ``torch.nn`` or a function from
            `torch.nn.functional``, depending on whether it needs to be further
            parameterized or not. Overwrites `activate` of the current instance
            if given. Defaults to ``None``.
        gate: Module or function, optional
            The activation function to be applied to half of the (linearly)
            transformed input before multiplying with the other half. Must be
            a callable that accepts a tensor as sole argument, like a module
            from ``torch.nn`` or a function from ``torch.nn.functional``.
            Overwrites the `gate` of the current instance if given.
            Defaults to ``None``.
        drop: Module, optional
            Typically an instance of ``Dropout`` or ``AlphaDropout``.
            Overwrites the `drop` of the current instance if given.
            Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layers together.

        Returns
        -------
        GatedResidualConcatMixer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.n_features if n_features is None else n_features,
            self.activate if activate is None else activate,
            self.gate if gate is None else gate,
            self.drop if drop is None else drop,
            **(self.kwargs | kwargs)
        )



import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any
from pandas import DataFrame
import pandas as pd
from google.cloud import storage as gcs
from ...misc import ArgRepr


class GcsParquet2DataFrame(ArgRepr):
    """Load parquet files from Google Cloud Storage into a pandas dataframe.

    Parameters
    ----------
    project: str
        Project where the `bucket` and parquet files reside.
    bucket: str
        Bucket where the parquet files reside.
    prefix: str, optional
        The prefix of the parquet files to download. Since it (or part of it)
        can also be provided later, when the callable instance is called, it
        is optional here. Defaults to an empty string.
    n_threads: int, optional
        Maximum number of parquet files to download in parallel.
        Defaults to 16.
    chunk_size: int, optional
        Chunk size to read from Google Cloud Storage in one API call in MiB.
        Defaults to 10 MiB.
    **kwargs
        Additional keyword arguments are passed to the constructor of the
        Google Storage ``Client`` (see `documentation <https://cloud.google.
        com/python/docs/reference/storage/latest/google.cloud.storage.
        client.Client#parameters>`__ for options).

    """

    __thread = threading.local()

    def __init__(
            self,
            project: str,
            bucket: str,
            prefix: str = '',
            n_threads: int = 16,
            chunk_size: int = 10,
            **kwargs: Any
    ) -> None:
        self.project = project.strip(' /.')
        self.bucket = bucket.strip(' /.')
        self.prefix = prefix.strip(' ./') + '/' if prefix.strip(' ./') else ''
        self.n_threads = n_threads
        self.chunk_size = chunk_size
        self.kwargs = kwargs
        super().__init__(
            self.project,
            self.bucket,
            self.prefix,
            self.n_threads,
            self.chunk_size,
            **kwargs
        )

    @property
    def chunk_bytes(self) -> int:
        """Bytes to read from Google Cloud Storage in one API call."""
        in_bytes = self.chunk_size * 1024 * 1024
        in_multiples_of_256kb = int(in_bytes // (256 * 1024))
        return in_multiples_of_256kb * 256 * 1024

    def __call__(self, prefix: str = '') -> DataFrame:
        """Load parquet files from Google Cloud Storage into pandas DataFrame.

        Parameters
        ----------
        prefix: str, optional
            The prefix of the parquet files to load. If given here, it will
            be appended to the `prefix` given at instantiation time.
            Defaults to an empty string.

        Returns
        -------
        DataFrame
            Concatenated parquet files.

        """
        prefix = prefix.strip(' ./') + '/' if prefix.strip(' ./') else ''
        remote = self.prefix + prefix

        client = gcs.Client(self.project, **self.kwargs)
        blobs = client.list_blobs(self.bucket, prefix=remote or None)

        with ThreadPoolExecutor(
                self.n_threads,
                initializer=self.__initializer
        ) as pool:
            downloads = as_completed(
                pool.submit(self.__download, blob.name)
                for blob in blobs
                if blob.name.count('/') == remote.count('/')
            )
            results = [download.result() for download in downloads]

        df = pd.concat(results, copy=False) if results else DataFrame()
        df.reset_index(drop=True, inplace=True)
        return df

    def __initializer(self) -> None:
        """Each thread needs its own client as they cannot be pickled."""
        self.__thread.client = gcs.Client(self.project, **self.kwargs)
        self.__thread.bucket = self.__thread.client.get_bucket(self.bucket)

    def __download(self, name: str) -> DataFrame:
        """Read parquet from binary blob opened as file-like object."""
        blob = self.__thread.bucket.get_blob(name)
        with blob.open(
                'rb',
                chunk_size=self.chunk_bytes,
                raw_download=True
        ) as stream:
            df = pd.read_parquet(stream, buffer_size=self.chunk_bytes)
        return df



"""Tools to interact with hosted cloud services.

Current only supports elements of the Google Cloud Project (GCP) but, in
the future, Amazon Web Services (AWS) might get implemented as well.

"""

from importlib.util import find_spec

required = 'google_cloud_bigquery', 'google_cloud_storage', 'pandas_gbq'

if any(find_spec(package) for package in required) is None:
    msg = 'Install {} with the [cloud] extra to unlock this subpackage!'
    raise ImportError(msg.format(__package__.split('.')[0]))



from collections.abc import Iterable
from ..misc import ArgRepr
from .exceptions import SumError


class Sum[S, T](ArgRepr):
    """Equivalent to a partial of the python builtin ``sum`` function.

    Upon subclassing and/or instantiation, type annotation with the type of
    the elements in the iterable and the type of the final sum is recommended.

    Parameters
    ----------
    acc: optional
        The initial value of the sum to which all other elements are added.
        If not given, calling sum on an empty iterable will fail.
        Defaults to ``None``.

    """

    def __init__(self, acc: T | None = None) -> None:
        super().__init__(acc)
        self.acc = acc

    def __call__(self, iterable: Iterable[S]) -> T:
        """Sum up the elements of an iterable.

        Parameters
        ----------
        iterable: Iterable
            An iterable of elements to sum up.

        Returns
        -------
        object
            The sum of the elements in the `iterable`.

        """
        iterator = iter(iterable)
        acc = next(iterator) if self.acc is None else self.acc
        offset = 1 if self.acc is None else 0
        for i, element in enumerate(iterator):
            try:
                acc = acc + element
            except Exception as error:
                msg = 'Error adding element #{}:\n{}\n{}:\n{}'
                err_cls = error.__class__.__name__
                fmt = msg.format(i + offset, element, err_cls, error)
                raise SumError(fmt) from error
        return acc



from typing import Any, Self
import torch.nn as ptn
from ..types import Tensor, Module, Functional, Drop


class GatedResidualEmbedder(Module):
    """Gated Residual Network (GRN) for embedding numerical features.

    Parameters
    ----------
    mod_dim: int
        Desired embedding size. Will become the size of the last dimension of
        the output tensor.
    activate: Module or function, optional
        The activation function to be applied after (linear) projection into
        embedding space, but prior to gating. Must be a callable that accepts
        a tensor as sole argument, like a module from ``torch.nn`` or a
        function from `torch.nn.functional``, depending on whether it needs to
        be further parameterized or not. Defaults to an ``ELU`` activation.
    gate: Module or function, optional
        The activation function to be applied to half of the (non-linearly)
        projected input before multiplying with the other half. Must be
        a callable that accepts a tensor as sole argument, like a module from
        ``torch.nn`` or a function from ``torch.nn.functional``, depending
        on whether it needs to be further parameterized or not.
        Defaults to a sigmoid.
    drop: Module, optional
        Typically an instance of ``Dropout`` or ``AlphaDropout``. Defaults to
        ``Dropout(p=0.0)``, resulting in no dropout being applied.
    inp_dim: int, optional
        The number of features to embed together. Defaults to 1.
    **kwargs
        Additional keyword arguments to pass through to the linear layers.

    Notes
    -----
    This implementation is inspired by how features are encoded in `Temporal
    Fusion Transformers`, [1]_ but it is not quite the same. Firstly, the
    (linear) projection of scalar numerical features into embedding space
    happens inside the present module. Secondly, this embedding vector is not
    transformed again (as Eq. 4 seems to imply) and there is no option to add
    a `context vector`. Thirdly, the intermediate linear layer (Eq. 3) is
    eliminated and dropout is applied directly to the activations after the
    first layer. Finally, the layer norm (Eq. 2) is replaced by simply
    dividing the sum of (linearly projected) input and gated signal by 2.
    Should additional normalization be desired, it can be performed
    independently on the output of this module.

    References
    ----------
    .. [1] B. Lim, S. O. Arik, N. Loeff, and T. Pfister, `Temporal Fusion
           Transformers for Interpretable Multi-horizon Time Series
           Forecasting`, `arXiv:1912.09363v3 <https://arxiv.org/abs/
           1912.09363>`__ (2020).

    """

    def __init__(
            self,
            mod_dim: int,
            activate: Module | Functional = ptn.ELU(),
            gate: Module | Functional = ptn.Sigmoid(),
            drop: Drop = ptn.Dropout(0.0),
            inp_dim: int = 1,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.activate = activate
        self.gate = gate
        self.drop = drop
        self.inp_dim = inp_dim
        self.kwargs = kwargs
        self.project = ptn.Linear(inp_dim, mod_dim, **kwargs)
        self.widen =  ptn.Linear(mod_dim, 2 * mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Embed a numerical feature through a Gated Residual Network (GRN).

        Parameters
        ----------
        inp: Tensor
            The last dimension of the input tensor is typically expected to be
            of size 1 and to contain the numerical value of a single feature.
            In case `inp_dim` dim was explicitly set to a value > 1 on
            instantiation, the size of the last dimension must match `inp_dim`,
            the number of numerical features to embed together.

        Returns
        -------
        Tensor
            The output has the same number of dimensions as the input with
            the size of the last dimension changed to the specified `mod_dim`.

        """
        projected = self.project(inp)
        wide = self.widen(self.drop(self.activate(projected)))
        gated = wide[..., :self.mod_dim] * self.gate(wide[..., self.mod_dim:])
        return 0.5 * (projected + gated)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.project.reset_parameters()
        self.widen.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            activate: Module | Functional | None = None,
            gate: Module | Functional | None = None,
            drop: Drop | None = None,
            inp_dim: int | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Desired embedding size. Will become the size of the last dimension
            of the output tensor. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        activate: Module or function, optional
            The activation function to be applied after (linear) projection
            into embedding space, but prior to gating. Must be a callable that
            accepts a tensor as sole argument, like a module from ``torch.nn``
            or a function from `torch.nn.functional``, depending on whether it
            needs to be further parameterized or not. Overwrites the `activate`
            of the current instance if given. Defaults to ``None``.
        gate: Module or function, optional
            The activation function to be applied to half of the (non-linearly)
            projected input before multiplying with the other half. Must be
            a callable that accepts a tensor as sole argument, like a module
            from ``torch.nn`` or a function from ``torch.nn.functional``.
            Overwrites the `gate` of the current instance if given.
            Defaults to ``None``.
        drop: Module, optional
            Typically an instance of ``Dropout`` or ``AlphaDropout``.
            Overwrites the `drop` of the current instance if given.
            Defaults to ``None``.
        inp_dim: int, optional
            The number of features to embed together. Overwrites the `inp_dim`
            of the current instance if given. Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layers together.

        Returns
        -------
        GatedResidualEmbedder
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.activate if activate is None else activate,
            self.gate if gate is None else gate,
            self.drop if drop is None else drop,
            self.inp_dim if inp_dim is None else inp_dim,
            **(self.kwargs | kwargs)
        )



from collections.abc import Callable, Iterable
from ..misc import ArgRepr
from .exceptions import SafeError


class Safe[**P, T](ArgRepr):
    """Wrap callable to catch potential errors and safely return them.

    Generic type annotation of instances is recommended. Provide a list of
    one or more input types that the callable takes, followed by the return
    type of the callable.

    Parameters
    ----------
    call: callable
        Callable to wrap.
    exception: optional
        Specific exception to catch (or an iterable of exceptions).
        Defaults to ``Exception``.
    *exceptions
        Additional exceptions to catch.

    See Also
    --------
    SafeError

    """

    def __init__(
            self,
            call: type[T] | Callable[P, T],
            exception: type[Exception] | Iterable[type[Exception]] = (),
            *exceptions: type[Exception]
    ) -> None:
        self.call = call
        try:
            exception = tuple(exception)
        except TypeError:
            exception = exception,
        self.exceptions = tuple(set(exception + exceptions))
        super().__init__(call, *self.exceptions)
        self.exceptions = self.exceptions if self.exceptions else (Exception,)

    def __call__(self, *args: P.args) -> T | SafeError:
        """Call the cached callable, catching any or all exceptions raised.

        Parameters
        ----------
        *args
            Arguments to call `call` with.

        Returns
        -------
        object or SafeError
            Either the return value(s) of the cached `call` or an instance of
            ``SafeError`` wrapping one of the specified `exceptions`.

        """
        try:
            return self.call(*args)
        except self.exceptions as error:
            return SafeError(error, self._name(self.call), self.call, args)



import pkgutil
import warnings

from ..misc import ArgRepr
from .misc import NotFound


class TextResourceLoader(ArgRepr):
    """Load text files from a resource directory within a python package.

    Parameters
    ----------
    package: str
        Name of the python package under which the text file is located.
    path: str, optional
        Directory under which the text file is located within the python
        package or full path to the text file. If not fully specified here,
        the path must be completed on calling the instance.
        Defaults to "resources".
    not_found: str, optional
        What to do if the specified file is not found. One of "ignore", "warn",
        or "raise". Defaults to "raise". Use the ``NotFound`` enum
        to avoid typos!
    encoding: str, optional
        Encoding of the text file. Defaults to "utf-8".

    See Also
    --------
    NotFound

    """

    def __init__(
            self,
            package: str,
            path: str = 'resources',
            not_found: str = NotFound.RAISE,
            encoding: str = 'utf-8'
    ) -> None:
        self.package = package.strip(' ./')
        self.path = path.strip(' /')
        self.not_found = not_found.strip().lower()
        self.encoding = encoding.strip()
        super().__init__(
            self.package,
            self.path,
            self.not_found,
            self.encoding
        )

    def __call__(self, path: str = '') -> str:
        """Load text file from a directory within the specified python package.

        Parameters
        ----------
        path: str, optional
            Path (including file name) relative to the `path` specified at
            instantiation. Defaults to an empty string, which results in an
            unchanged `path` on concatenation.

        Returns
        -------
        str
            Decoded contents of the specified text file.

        """
        path = '/' + path.strip(' /') if path.strip(' /') else ''
        full_path = self.path + path
        try:
            content = pkgutil.get_data(self.package, full_path)
        except FileNotFoundError as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'File {} not found!\nReturning empty string.'
                    warnings.warn(msg.format(full_path))
                    content = b''
                case NotFound.IGNORE:
                    content = b''
                case _:
                    raise error
        return content.decode(self.encoding)



from typing import Any
from collections.abc import Callable
from ..misc import ArgRepr


class Partial[T](ArgRepr):
    """Alternative implementation of ``functools.partial``.

    Positional arguments given when instances are called are *appended* to
    those given at instantiation . Upon subclassing and/or instantiation,
    annotation with the (return) type of `call` is recommended.

    Parameters
    ----------
    call: callable
        Callable object with arbitrary call signature.
    *args
        Any number of arguments to call `call` with.
    **kwargs
        Any number of keyword arguments to call `call` with.

    See Also
    --------
    Curry

    """

    def __init__(
            self,
            call: type[T] | Callable[..., T],
            *args: Any,
            **kwargs: Any
    ) -> None:
        super().__init__(call, *args, **kwargs)
        self.call = call
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        """Execute cached callable with combined args and kwargs.

        Parameters
        ----------
        *args
            Additional arguments are appended to cached arguments.
        **kwargs
            Additional keyword arguments are combined with cached keyword
            arguments, overriding them in case of conflict.

        Returns
        -------
        object
            Whatever `call` returns.

        """
        return self.call(*(*self.args, *args), **{**self.kwargs, **kwargs})



import os
import json
from json import JSONDecodeError
from ast import literal_eval
from typing import Any
from ..misc import ArgRepr


class EnvParser(ArgRepr):
    """Parse OS environment variables, preferring prefixed over pure versions.

    Sometimes, environment variables desired for individual use are already
    taken by the operating system or some other system component. In these
    cases, one can resort to prefixing these to avoid conflicts. The present
    class is instantiated with that prefix and will resolve conflicts when
    objects are called, returning the OS environment as a dictionary with
    the values of all variables parsed into python literals.

    Parameters
    ----------
    prefix: str, optional
        Prefix of environment variables that would otherwise be shadowed
        by existing ones. Defaults to empty string.

    """

    def __init__(self, prefix: str = '') -> None:
        super().__init__(prefix)
        self.prefix = prefix

    def __call__(self, env: dict[str, str] | None = None) -> dict[str, Any]:
        """Parse the OS environment, resolving potentially prefixed variables.

        Parameters
        ----------
        env: dict, optional
            Dictionary to be parsed and resolved. Defaults to ``os.environ``.

        Returns
        -------
        dict
            Environment with prefixed keys removed and the values of their
            non-prefixed counterparts updated accordingly.

        """
        env = os.environ if env is None else env
        prefixed = {}
        original = {}
        for key in env:
            if key.startswith(self.prefix):
                prefixed[key.removeprefix(self.prefix)] = env[key]
            else:
                original[key] = env[key]
        merged = original | prefixed
        return {key: self.__parsed(value) for key, value in merged.items()}

    @staticmethod
    def __parsed(value: str) -> Any:
        """Try to parse (string) environment variables into python objects."""
        try:
            parsed = json.loads(value)
        except (TypeError, JSONDecodeError):
            try:
                parsed = literal_eval(value)
            except (TypeError, ValueError, SyntaxError):
                parsed = value
        return parsed



import warnings
import tomllib
from typing import Any
from collections.abc import Callable
from pathlib import Path
import yaml
from yaml import Loader
from ..misc import ArgRepr
from .misc import NotFound

type Toml = dict[str, Any]
type Yaml = dict[str, Any] | list[Any]


class TomlReader(ArgRepr):
    """Light wrapper around the python standard library's ``tomllib.load``.

    Parameters
    ----------
    path: str, optional
        Directory under which the TOML file is located or full path to the
        TOML file. If not fully specified here, the path must be completed on
        calling the instance. Defaults to the current working directory of the
        python interpreter.
    not_found: str, optional
        What to do if the specified TOML file is not found. One of "ignore",
        "warn", or "raise". Defaults to "raise". Use the ``NotFound`` enum to
        avoid typos!
    parse_float: callable, optional
        Will be called with the string of every TOML float to be decoded.
        Defaults to ``float``.
    **kwargs
        Additional keyword arguments will be forwarded python`s builtin
        ``open`` function. Since the `mode` must be ``"rb"``, it is dropped
        from the keyword arguments.

    See Also
    --------
    NotFound

    """

    def __init__(
            self,
            path: str = '',
            not_found: str = NotFound.RAISE,
            parse_float: Callable[[str], float] = float,
            **kwargs: Any
    ) -> None:
        self.path = path.strip()
        self.not_found = not_found.strip().lower()
        self.parse_float = parse_float
        if 'mode' in kwargs:
            self.kwargs = (kwargs.pop('mode'), kwargs)[1]
        else:
            self.kwargs = kwargs
        super().__init__(self.path, self.not_found, parse_float, **self.kwargs)

    def __call__(self, path: str = '') -> Toml:
        """Read a specific TOML file.

        If `not_found` is set to "warn" or "ignore" and the file cannot be
        found, an empty dictionary is returned.

        Parameters
        ----------
        path: str
            Path (including file name) relative to the `path` specified at
            instantiation. Defaults to an empty string, which results in an
            unchanged `path` on concatenation.

        Returns
        -------
        dict
            The parsed contents of the TOML file.

        """
        path = Path(self.path) / path.strip(' /')
        try:
            with path.open('rb', **self.kwargs) as file:
                toml = tomllib.load(file, parse_float=self.parse_float)
        except FileNotFoundError as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'File {} not found!\nReturning empty TOML.'
                    warnings.warn(msg.format(path.resolve()))
                    toml = {}
                case NotFound.IGNORE:
                    toml = {}
                case _:
                    raise error
        return toml


class YamlReader(ArgRepr):
    """Light wrapper around pyyaml's ``yaml.load`` function.

    Parameters
    ----------
    path: str, optional
        Directory under which the YAML file is located or full path to the
        YAML file. If not fully specified here, the path must be completed on
        calling the instance. Defaults to the current working directory of the
        python interpreter.
    not_found: str, optional
        What to do if the specified YAML file is not found. One of "ignore",
        "warn", or "raise". Defaults to "raise". Use the ``NotFound`` enum to
        avoid typos!
    loader: type, optional
        The loader class to use. Defaults to ``Loader``
    **kwargs
        Additional keyword arguments will be forwarded python`s builtin
        ``open`` function. The `mode` is hard-coded to ``"rb"`` and is dropped
        from the keyword arguments.

    See Also
    --------
    NotFound

    """

    def __init__(
            self,
            path: str = '',
            not_found: str = NotFound.RAISE,
            loader: type = Loader,
            **kwargs: Any
    ) -> None:
        self.path = path.strip()
        self.not_found = not_found.strip().lower()
        self.loader = loader
        if 'mode' in kwargs:
            self.kwargs = (kwargs.pop('mode'), kwargs)[1]
        else:
            self.kwargs = kwargs
        super().__init__(self.path, self.not_found, loader, **self.kwargs)

    def __call__(self, path: str = '') -> Yaml:
        """Read a specific YAML file.

        If `not_found` is set to "warn" or "ignore" and the file cannot be
        found, and empty dictionary is returned.

        Parameters
        ----------
        path: str
            Path (including file name) relative to the `path` specified at
            instantiation. Defaults to an empty string, which results in an
            unchanged `path` on concatenation.

        Returns
        -------
        dict or list
            The parsed contents of the YAML file.

        """
        path = Path(self.path) / path.strip(' /')
        try:
            with path.open('rb', **self.kwargs) as file:
                yml = yaml.load(file, self.loader)
        except FileNotFoundError as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'File {} not found!\nReturning empty YAML.'
                    warnings.warn(msg.format(path.resolve()))
                    yml = {}
                case NotFound.IGNORE:
                    yml = {}
                case _:
                    raise error
        return yml or {}



from typing import Any, Self
from collections.abc import Iterable
from functools import cached_property
import torch as pt
import torch.nn as ptn
from ..types import Tensor, Module


class CategoricalEmbedder(Module):
    """Embed one or more categorical features as numerical vectors.

    Parameters
    ----------
    mod_dim: int
        Desired embedding size. Will become the size of the last dimension of
        the output tensor.
    cat_count: int or iterable of int, optional
        One integer or an iterable (e.g., a tuple or list) of integers, each
        specifying the total number of categories in the respective feature.
        Defaults to an emtpy tuple.
    *cat_counts: int
        Category counts for additional features. Together with the `cat_count`,
        the total number of category counts, i.e., the total number of features
        to embed must match the size of the last dimension of the input tensor.
    **kwargs
        Keyword arguments are forwarded to the PyTorch ``Embedding`` class.

    Notes
    -----
    The integer numbers identifying a category are expected to be zero-base,
    i.e., if the category count of a feature is 3, the allowed category
    identifier are 0, 1, and 2. If you need a padding index (e.g., to mark
    missing/unknown values), do not forget to increase all `cat_counts` by one!

    """

    def __init__(
            self,
            mod_dim: int,
            cat_count: int | Iterable[int] = (),
            *cat_counts: int,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        cat_count = self.__valid(cat_count)
        self.cat_counts: tuple[int, ...] = cat_count + self.__valid(cat_counts)
        self.kwargs = kwargs
        self.embed = ptn.ModuleList([
            ptn.Embedding(count, mod_dim, **kwargs)
            for count in self.cat_counts
        ])

    @property
    def n_features(self) -> int:
        """Number of features to embed."""
        return len(self.cat_counts)

    @cached_property
    def features(self) -> range:
        """Range of feature indices."""
        return range(self.n_features)

    @cached_property
    def dim(self) -> int:
        """The output tensor dimension index to stack features into."""
        return -1 if self.n_features < 1 else -2

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass for embedding categorical features into vectors.

        Parameters
        ----------
        inp: tensor
            Input tensor must be of dtype ``long``. The size of the last
            dimension is expected to match the number of specified `cat_counts`
            and to contain the integer identifiers of the categories in the
            respective feature. These identifiers must all be lower in value
            than their respective count.

        Returns
        -------
        Tensor
            The output tensor has one more dimension of size `mod_dim` added
            after the last dimension (with a size equal to the number of
            `cat_counts`) than the `inp`, containing the stacked embeddings.

        """
        emb = [self.embed[cat](inp[..., cat]) for cat in self.features]
        return pt.stack(emb or self.mod_dim * [pt.zeros(*inp.shape)], self.dim)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        for emb in self.embed:
            emb.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            cat_count: int | Iterable[int] | None = None,
            *cat_counts: int,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Desired embedding size. Will become the size of the last dimension
            of the output tensor. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        cat_count: int or iterable of int, optional
            One integer or an iterable (e.g., tuple or list) of integers, each
            specifying the number of categories in the respective feature.
            Overwrites the `cat_count` of the current instance if given.
            Defaults to ``None``.
        *cat_counts: int
            Category counts for additional features. Together with the
            `cat_count`, the total number of category counts must match the
            size of the last dimension of the input tensor.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then used together for
            instantiating the PyTorch ``Embedding`` class.

        Returns
        -------
        CategoricalEmbedder
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.cat_counts if cat_count is None else cat_count,
            *cat_counts,
            **(self.kwargs | kwargs)
        )

    @staticmethod
    def __valid(cat_counts: int | Iterable[int]) -> tuple[int, ...]:
        """Ensure that the argument is indeed an iterable of integers."""
        try:
            return tuple(abs(int(cat_count)) for cat_count in cat_counts)
        except TypeError:
            return abs(int(cat_counts)),



class SchemaError(Exception):
    pass


class DefaultsError(Exception):
    pass


class ParseError(Exception):
    pass


class CastError(Exception):
    pass


class ValidationErrors(ExceptionGroup):
    pass



from collections.abc import Callable, Iterable
from ..misc import ArgRepr
from .exceptions import FilterError


class Filter[S, T](ArgRepr):
    """Equivalent to a partial of the python builtin ``filter`` function.

    Upon subclassing and/or instantiation, type annotation with the type of the
    elements in the iterator to be filtered and the return type of `wrapper`
    is recommended.

    Parameters
    ----------
    criterion: callable, optional
        Callable accepting one element of an iterable at a time and returning
        a boolean value for each one of them. Defaults to ``None``, which
        returns the inherent truth values of the object in the iterable.
    wrapper: type or callable, optional
        If not given, an attempt will be made to return the same type of
        iterable the callable instance is being called with (by calling its
        class with a list of the filtered elements). If explicitly given,
        `wrapper` will be called with a list of filtered elements.
        Consequently, the return type will be the (return) type of `wrapper`.

    Notes
    -----
    In contrast to python's builtin lazy ``filter`` function, which returns a
    generator object, the filtered iterable is fully manifested first and only
    then wrapped.

    """

    def __init__(
            self,
            criterion: Callable[[S], bool] | None = None,
            wrapper: type[T] | Callable[[list[S]], T] | None = None
    ) -> None:
        super().__init__(criterion, wrapper)
        self.criterion = criterion
        self.wrapper = wrapper

    def __call__(self, iterable: Iterable[S]) -> T:
        """Filter an iterable according to the specified criterion.

        Parameters
        ----------
        iterable: Iterable
            An iterable whose elements are to be filtered according to whether
            the `criterion` evaluates to ``True`` or ``False``.

        Returns
        -------
        Sequence
            Same type as `iterable` if `wrapper` was not specified on
            instantiation or the (return) type of `wrapper`.

        Raises
        ------
        FilterError
            If calling the criterion on any element of `iterable` raises
            an exception or if wrapping the results leads to an exception.

        """
        criterion = bool if self.criterion is None else self.criterion
        filtered = []
        for i, element in enumerate(iterable):
            try:
                criterion_is_fulfilled = criterion(element)
            except Exception as error:
                msg = '\n{} calling criterion\n{}\non element #{}:\n{}\n{}'
                name = self._name(criterion)
                err_cls = error.__class__.__name__
                fmt = msg.format(err_cls, name, i, element, error)
                raise FilterError(fmt) from error
            if criterion_is_fulfilled:
                filtered.append(element)
        wrap = iterable.__class__ if self.wrapper is None else self.wrapper
        try:
            wrapped = wrap(filtered)
        except Exception as err:
            msg = '\n{} calling wrapper\n{}\non filter results:\n{}'
            name = self._name(wrap)
            err_cls = err.__class__.__name__
            raise FilterError(msg.format(err_cls, name, err)) from err
        return wrapped



"""Flexible and composable building blocks for constructing neural-networks.

After features are embedded and combined, it is time to extract as much
information as possible to predict the desired target. One way of doing this
systematically is to repeat layers of identical internal architecture with
residual (or skip) connections between them.

"""

from typing import Any, Self
from abc import ABC, abstractmethod
import torch.nn as ptn
from .types import Module, Tensor, Functional, Drop
from .misc import Identity

__all__ = [
    'Block',
    'ActivatedBlock',
    'GatedBlock',
    'GatedResidualBlock',
    'SkipConnection',
    'Repeat'
]


class Block(Module, ABC):
    """Abstract base class for stackable/repeatable neural-network components.

    The input and output tensors of such components must have the same
    dimensions and sizes!

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    @abstractmethod
    def new(self) -> Self:
        """Return a fresh, new instance with exactly the same parameters."""
        ...

    @abstractmethod
    def reset_parameters(self) -> None:
        """Subclasses implement in-place reset of all internal parameters."""
        ...


class ActivatedBlock(Block):
    """A single, non-linearly activated hidden layer of configurable size.

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension and the output will again have this size in
        its last dimension.
    activate: Module or function, optional
        The activation function to be applied after projecting into higher-
        dimensional space. Must be a callable that accepts a tensor as sole
        argument, like a module from ``torch.nn`` or a function from
        ``torch.nn.functional``, depending on whether it needs to be further
        parameterized or not. Defaults to ``ELU()``.
    hidden_factor: int, optional
        The size of the hidden layer is this integer factor times `mod_dim`.
        Defaults to 4.
    **kwargs
        Additional keyword arguments to pass through to the linear layers.

    """

    def __init__(
            self,
            mod_dim: int,
            activate: Module | Functional = ptn.ELU(),
            hidden_factor: int = 4,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.activate = activate
        self.hidden_factor = hidden_factor
        self.kwargs = kwargs
        self.widen = ptn.Linear(mod_dim, hidden_factor * mod_dim, **kwargs)
        self.shrink = ptn.Linear(hidden_factor * mod_dim, mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass through a single, non-linearly activated hidden layer.

        Parameters
        ----------
        inp: Tensor
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        Tensor
            Same dimensions and sizes as the input tensor.

        """
        return self.shrink(self.activate(self.widen(inp)))

    def reset_parameters(self) -> None:
        """Re-initialize the internal parameters of the linear projections."""
        self.widen.reset_parameters()
        self.shrink.reset_parameters()

    def new(self) -> Self:
        """Return a fresh, new instance with exactly the same parameters."""
        return self.__class__(
            self.mod_dim,
            self.activate,
            self.hidden_factor,
            **self.kwargs
        )


class GatedBlock(Block):
    """A configurable, gated linear unit (GLU).

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension and the output will again have this size in
        its last dimension.
    gate: Module or function
        gate: Module or function, optional
        The activation function to be applied to half of the (linearly)
        projected input before multiplying with the other half. Must be
        a callable that accepts a tensor as sole argument, like a module from
        ``torch.nn`` or a function from ``torch.nn.functional``, depending
        on whether it needs to be further parameterized or not.
        Defaults to a sigmoid.
    **kwargs
        Additional keyword arguments to pass through to the linear layers.

    """

    def __init__(
            self,
            mod_dim: int,
            gate: Module | Functional = ptn.Sigmoid(),
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.gate = gate
        self.kwargs = kwargs
        self.widen = ptn.Linear(mod_dim, 2 * mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass through a single gated linear unit (GLU).

        Parameters
        ----------
        inp: Tensor
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        Tensor
            Same dimensions and sizes as the input tensor.

        """
        wide = self.widen(inp)
        return wide[..., :self.mod_dim] * self.gate(wide[..., self.mod_dim:])

    def reset_parameters(self) -> None:
        """Re-initialize the internal parameters of the linear projection."""
        self.widen.reset_parameters()


    def new(self) -> Self:
        """Return a fresh, new instance with exactly the same parameters."""
        return self.__class__(
            self.mod_dim,
            self.gate,
            **self.kwargs
        )


class GatedResidualBlock(Block):
    """Gated Residual Network (GRN) for efficiently extracting information.

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension and the output will again have this size in
        its last dimension.
    activate: Module or function, optional
        The activation function to be applied after (linear) projection, but
        prior to gating. Must be a callable that accepts a tensor as sole
        argument, like a module from ``torch.nn`` or a function from
        `torch.nn.functional``, depending on whether it needs to be further
        parameterized or not. Defaults to an ``ELU`` activation.
    gate: Module or function, optional
        The activation function to be applied to half of the (non-linearly)
        projected input before multiplying with the other half. Must be
        a callable that accepts a tensor as sole argument, like a module from
        ``torch.nn`` or a function from ``torch.nn.functional``, depending
        on whether it needs to be further parameterized or not.
        Defaults to a sigmoid.
    drop: Module, optional
        Dropout to be applied within the GRN. Typically an instance of
        ``Dropout`` or ``AlphaDropout``. Defaults to ``Dropout(p=0.0)``,
        resulting in no dropout being applied.
    **kwargs
        Additional keyword arguments to pass through to the linear layers.

    Notes
    -----
    This implementation is inspired by how features are encoded in `Temporal
    Fusion Transformers`, [1]_ but it is not quite the same. Firstly, the
    intermediate linear layer (Eq. 3) is eliminated and dropout is applied
    directly to the activations after the first layer. Secondly, the layer norm
    (Eq. 2) is replaced by simply dividing the sum of (linearly projected)
    input and gated signal by 2. Should additional normalization be desired, it
    can be performed independently on the output of this module.

    References
    ----------
    .. [1] B. Lim, S. O. Arik, N. Loeff, and T. Pfister, `Temporal Fusion
           Transformers for Interpretable Multi-horizon Time Series
           Forecasting`, `arXiv:1912.09363v3 <https://arxiv.org/abs/
           1912.09363>`__ (2020).

    """

    def __init__(
            self,
            mod_dim: int,
            activate: Module | Functional = ptn.ELU(),
            gate: Module | Functional = ptn.Sigmoid(),
            drop: Drop = ptn.Dropout(0.0),
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.activate = activate
        self.gate = gate
        self.drop = drop
        self.kwargs = kwargs
        self.project = ptn.Linear(mod_dim, mod_dim, **kwargs)
        self.widen = ptn.Linear(mod_dim, 2 * mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass through a single gated residual network (GRN).

        Parameters
        ----------
        inp: Tensor
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        Tensor
            Same dimensions and sizes as the input tensor.

        """
        projected = self.project(inp)
        wide = self.widen(self.drop(self.activate(projected)))
        gated = wide[..., :self.mod_dim] * self.gate(wide[..., self.mod_dim:])
        return 0.5 * (projected + gated)

    def reset_parameters(self) -> None:
        """Re-initialize the internal parameters of the linear projections."""
        self.project.reset_parameters()
        self.widen.reset_parameters()

    def new(self) -> Self:
        """Return a fresh, new instance with exactly the same parameters."""
        return self.__class__(
            self.mod_dim,
            self.activate,
            self.gate,
            self.drop,
            **self.kwargs
        )


class SkipConnection(Block):
    """Add a residual/skip connection around the wrapped neural-network block.

    Parameters
    ----------
    block: Block
        The block to wrap the residual/skip connection around. The reason why
        this cannot simply be a ``Module`` is that currently PyTorch does not
        provide a reasonable way of cloning them.
    drop: Module, optional
        Dropout to be applied to the output of `block` before adding it to
        its input. Typically an instance of ``Dropout`` or ``AlphaDropout``.
        Defaults to ``Dropout(p=0.0)``, resulting in no dropout being applied.
    norm_cls: type, optional
        The class of the norm to be applied after adding input to output, e.g.,
        ``LayerNorm`` or ``BatchNorm1d``. Again, this is needed to easily
        create a fresh, new instances with equal, but independent parameters.
        Defaults to ``Identity``, resulting in no normalization whatsoever.
    *args
        Arguments used to initialize and instance of `norm_cls`.
    **kwargs
        Keyword arguments used to initialize and instance of `norm_cls`.

    """

    def __init__(
            self,
            block: Block,
            drop: Drop = ptn.Dropout(0.0),
            norm_cls: type[Module] = Identity,
            *args,
            **kwargs
    ) -> None:
        super().__init__()
        self.block = block
        self.drop = drop
        self.norm_cls: type[Module] = norm_cls
        self.args = args
        self.kwargs = kwargs
        self.norm = norm_cls(*args, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass through a block with the input added to the output.

        Parameters
        ----------
        inp: Tensor
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        Tensor
            Same dimensions and sizes as the input tensor.

        """
        return self.norm(0.5 * (inp + self.drop(self.block(inp))))

    def reset_parameters(self) -> None:
        """Re-initialize the internal parameters of the block and the norm."""
        self.block.reset_parameters()
        self.norm.reset_parameters()

    def new(self) -> Self:
        """Return a fresh, new instance with exactly the same parameters."""
        return self.__class__(
            self.block.new(),
            self.drop,
            self.norm_cls,
            *self.args,
            **self.kwargs
        )


class Repeat(Block):
    """Repeat a skip-connection, distilling ever finer detail from your data.

    Parameters
    ----------
    skip: SkipConnection
        An instance of a ``SkipConnection`` to repeat.
    n_layers: int, optional
        How often to repeat the `skip`. Defaults to 2.

    """

    def __init__(self, skip: SkipConnection, n_layers: int = 2) -> None:
        super().__init__()
        self.skip = skip
        self.n_layers = n_layers
        self.sequence = ptn.Sequential(*[skip.new() for _ in self.layers])

    @property
    def layers(self) -> range:
        """Range of layer indices."""
        return range(self.n_layers)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass through a stack of identical skip-connection blocks.

        Parameters
        ----------
        inp: Tensor
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        Tensor
            Same dimensions and sizes as the input tensor.

        """
        return self.sequence(inp)

    def reset_parameters(self) -> None:
        """Re-initialize the internal parameters of all blocks."""
        for block in self.sequence:
            block.reset_parameters()

    def new(self) -> Self:
        """Return a fresh, new instance with exactly the same parameters."""
        return self.__class__(
            self.skip,
            self.n_layers
        )



from typing import Any, Self
from collections.abc import Iterator, Callable, Sequence, Iterable
from functools import singledispatchmethod
from ..misc import IndentRepr
from .exceptions import RouteError

type Call = type | Callable[..., Any]
type Calls = tuple[Call, ...]
type Routes = Sequence[int | Sequence[int]]


class Route[**P, T](IndentRepr):
    """Flexibly route arguments to a sequence of callables and collect results.

    Generic type annotation of instances is recommended. Provide a list of
    one or more input types that will be routed to the callables and a
    ``tuple`` specifying the concatenation of the return types of all
    callables,  ignoring empty tuples. If only a single object remains, the
    type of that object should be annotated.

    Parameters
    ----------
    routes: sequence of int or sequence of sequences of int, optional
        Specified as, e.g., ``[2, 0, 1]`` means that the first callable will be
        called with the third argument (index 2), the second with the first,
        and the third with the second. If callables take more than one
        argument, `routes` can be specified as ``[(2, 0), (), 1]``, which
        means that the first callable will be called with the third and first
        arguments, the second with no arguments, and the third with the second
        argument. Defaults to an empty tuple, meaning that no callables can be
        specified and, that, therefore, nothing is returned when calling the
        instance, no matter how many arguments it is called with.
    call: callable or iterable of callables, optional
        One callable or an iterator of callables that will be called with
        the arguments according to `routes`. Defaults to an empty tuple.
    *calls: callable
        Additional callables that will be called with the arguments according
        to `routes`. Together with `call`, there must be the same number
        of callables as there are routes.

    Raises
    ------
    RouteError
        If the `routes` cannot be parsed, if the number of `routes` does not
        match the number of callables specified with `call` and `calls`, or
        if (any of) `call` or any of `calls` is not, in fact, callable.

    """

    def __init__(
            self,
            routes: Routes = (),
            call: Call | Iterable[Call] = (),
            *calls: Call
    ) -> None:
        self.routes = self.__packed(routes)
        callables = self.__valid(call) + self.__valid(calls)
        self.calls = self.__compatible(*callables)
        routes = [r[0] if len(r) == 1 else r for r in self.routes]
        super().__init__(self.calls, routes)

    def __iter__(self) -> Iterator[Call]:
        # We could also iterate over instances of self ...
        return self.calls.__iter__()

    def __len__(self) -> int:
        return self.calls.__len__()

    def __bool__(self) -> bool:
        return bool(self.calls)

    def __contains__(self, item: Call) -> bool:
        return item in self.calls

    def __reversed__(self) -> Self:
        return self.__class__(
            list(reversed(self.routes)),
            *reversed(self.calls)
        )

    @singledispatchmethod
    def __getitem__(self, index: int) -> Call:
        # We could also return instances of self ...
        return self.calls[index]

    @__getitem__.register
    def _(self, index: slice) -> Self:
        return self.__class__(self.routes[index], *self.calls[index])

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls == other.calls and self.routes == other.routes
        return NotImplemented

    def __ne__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.calls != other.calls or self.routes != other.routes
        return NotImplemented

    def __add__(self, other: Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(
                [*self.routes, *other.routes],
                *self.calls, *other.calls
            )
        return NotImplemented

    def __call__(self, *args: P.args) -> T:
        """Distribute arguments according to `routes` and forward to `calls`.

        Parameters
        ----------
        *args
            Arguments to be redistributed according to `routes` among `calls`.
            There must be at least `n_args` arguments. Extras will be ignored.

        Returns
        -------
        tuple or object
            Concatenation of all return values of all `calls` in order. If only
            one of the `calls` returns something other than an empty tuple,
            that object is returned.

        Raises
        ------
        RouteError
            If there are too few arguments to redistribute among `calls`
            according to `routes` or if calling one of the `calls` fails.

        """
        if (n_args := len(args)) < self.n_args:
            msg = 'Number of arguments must be at least {}, not {}!'
            raise RouteError(msg.format(self.n_args, n_args))
        results = []
        for i, (route, call) in enumerate(zip(self.routes, self)):
            try:
                result = call(*tuple(args[r] for r in route))
            except Exception as error:
                msg = '\n{} executing\n{}\nin route #{} {} of\n{}\n{}'
                err_cls = error.__class__.__name__
                name = self._name(call)
                fmt = msg.format(err_cls, name, i, route, self, error)
                raise RouteError(fmt) from error
            else:
                if isinstance(result, tuple):
                    results.extend(result)
                else:
                    results.append(result)
        return results[0] if len(results) == 1 else tuple(results)

    @property
    def n_args(self) -> int:
        """The minimum number of arguments required for calling instances."""
        # Routes could be an empty tuple.
        if self.routes:
            # Each route could be an empty tuple.
            maxima = [max(route) for route in self.routes if route]
            # Maximum of the maximum integer in each route, if it exists.
            return max(maxima) + 1 if maxima else 0
        return 0

    @staticmethod
    def __valid(calls: Call | Iterable[Call]) -> tuple[Call, ...]:
        """Ensure that the argument is indeed an iterable of callables."""
        if callable(calls):
            return calls,
        iterable = True
        all_callable = False
        try:
            all_callable = all(callable(call) for call in calls)
        except TypeError:
            iterable = False
        if iterable and all_callable:
            return tuple(calls)
        raise RouteError('All paths in the route must be callable!')

    def __compatible(self, *calls: Call) -> Calls:
        """Check if the number of routes matches the number of callables."""
        if (n_routes := len(self.routes)) == (n_calls := len(calls)):
            return calls
        msg = 'Number of callables (={}) must match number of routes (={})!'
        raise RouteError(msg.format(n_calls, n_routes))

    @staticmethod
    def __packed(routes: Routes) -> tuple[tuple[int, ...], ...]:
        """Unify route specifications to tuples of integers."""
        tuples = []
        msg = 'Routes must be integers or tuples thereof, not {}'
        # Routes could be of completely wrong type, i.e., not iterable.
        try:
            safe_routes = [*routes]
        except TypeError as error:
            raise RouteError(msg.format(routes)) from error
        # If routes are iterable, elements could be ...
        for route in safe_routes:
            try:
                # ... some sort of sequence of integers ...
                tuples.append(tuple(int(r) for r in route))
            except TypeError:
                # ... or just a single integer.
                try:
                    tuples.append((int(route),))
                except TypeError as error:
                    raise RouteError(msg.format(route)) from error
            except ValueError as error:
                raise RouteError(msg.format(route)) from error
        return tuple(tuples)



from collections.abc import Callable, Iterator
from pandas import DataFrame
from torch import Tensor, dtype, device
from torch.nn import Module, Dropout, AlphaDropout
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

type Functional = Callable[[Tensor], Tensor]
type Tensors1T = tuple[Tensor]
type Tensors2T = tuple[Tensor, Tensor]
type Tensors3T = tuple[Tensor, Tensor, Tensor]
type Tensors4T = tuple[Tensor, Tensor, Tensor, Tensor]
type Tensors = tuple[Tensor, ...]
type Dtype = dtype
type Device = device
type Drop = Dropout | AlphaDropout
type Batch = tuple[Tensors, Tensor]
type Batches = Iterator[Batch]


__all__ = [
    'DataFrame',
    'Tensor',
    'Module',
    'Functional',
    'Tensors1T',
    'Tensors2T',
    'Tensors3T',
    'Tensors4T',
    'Tensors',
    'Dtype',
    'Device',
    'Drop',
    'Optimizer',
    'LRScheduler',
    'Batch',
    'Batches'
]



from operator import itemgetter
from typing import Self
from collections.abc import Hashable, Mapping, Callable, Iterator, Iterable
from functools import singledispatchmethod
from ..misc import ArgRepr


class ValuesGetter[T](ArgRepr):
    """Extract a sequence of dictionary values according to the given `keys`.

    Instances behave like sequences of these `keys` and can be interacted with
    accortdingly (e.g., one can index, slice, or reverse instances).

    Parameters
    ----------
    key: Hashable, optional
        Dictionary key or iterable thereof. Defaults to an empy tuple.
    *keys: Hashable
        Additional dictionary keys.
    wrapper: callable, optional
        A type or some other callable that transforms a tuple of dictionary
        values into some other type of sequence. The type of the sequence
        should be annotated. Instances can be type annotated with the type
        of sequence to be returned. Defaults to ``list``.

    Raises
    ------
    TypeError
        If (any of) `key` or any of `keys` are not, in fact, hashable.

    Notes
    -----
    Generic type annotation with the (return) type of `wrapper` is recommended.

    Examples
    --------
    >>> get_values = ValuesGetter('name', 'age')
    >>> get_values({'name': 'John', 'weight': 82, 'age': 32})
    ['John', 32]

    """

    def __init__(
            self,
            key: Hashable | Iterable[Hashable] = (),
            *keys: Hashable,
            wrapper: type[T] | Callable[[tuple], T] = list
    ) -> None:
        key = self.__valid(key)
        self.keys: tuple[Hashable, ...] = key + self.__valid(keys)
        self.wrapper = wrapper
        super().__init__(*self.keys, wrapper=wrapper)

    def __len__(self) -> int:
        return self.keys.__len__()

    def __bool__(self) -> bool:
        return bool(self.keys)

    def __reversed__(self) -> Self:
        return self.__class__(*reversed(self.keys))

    def __iter__(self) -> Iterator[Self]:
        return iter(self.__class__(key) for key in self.keys)

    def __contains__(self, key: Hashable) -> bool:
        return key in self.keys

    @singledispatchmethod
    def __getitem__(self, index: int) -> Self:
        return self.__class__(self.keys[index])

    @__getitem__.register
    def _(self, index: slice) -> Self:
        return self.__class__(*self.keys[index])

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.keys == other.keys
        return NotImplemented

    def __ne__(self, other: Self) -> bool:
        if isinstance(other, self.__class__):
            return self.keys != other.keys
        return NotImplemented

    def __add__(self, other: Hashable | Iterable[Hashable] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(self.keys, *other.keys)
        try:
            others = self.__valid(other)
        except TypeError:
            return NotImplemented
        return self.__class__(self.keys, *others)

    def __radd__(self, other: Hashable | Iterable[Hashable] | Self) -> Self:
        if isinstance(other, self.__class__):
            return self.__class__(other.keys, *self.keys)
        try:
            others = self.__valid(other)
        except TypeError:
            return NotImplemented
        return self.__class__(others, *self.keys)

    def __call__(self, mapping: Mapping) -> T:
        """Extract sequence of dictionary values for given keys.

        Parameters
        ----------
        mapping: Mapping
            Dictionary to extract values from.

        Returns
        -------
        Sequence
            The values extracted from the dictionary. The type of sequence
            is determined by the return type of `wrap`. Defaults to ``list``.

        """
        match len(self):
            case 0:
                values = ()
            case 1:
                values = itemgetter(*self.keys)(mapping),
            case _:
                values = itemgetter(*self.keys)(mapping)
        return self.wrapper(values)

    @staticmethod
    def __valid(keys: Hashable | Iterable[Hashable]) -> tuple[Hashable, ...]:
        """Ensure that the argument is indeed an iterable of hashables."""
        if isinstance(keys, str):
            return keys,
        try:
            _ = [hash(key) for key in keys]
        except TypeError:
            _ = hash(keys)
            return keys,
        return tuple(keys)



from typing import Any, Self
import torch.nn as ptn
from ..types import Tensor, Module, Functional
from ..misc import identity


class ActivatedConcatMixer(Module):
    """Combined stacked feature vectors through a single dense layer.

    Multiple feature vectors stacked in a single tensor are concatenated into a
    single, wide vector and projected down into a space the size of the model.

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension and the output will again have this size in
        its last dimension.
    n_features: int
        The number of features to combine. Must be equal to the size of the
        next-to-last dimension of the input tensor.
    activate: Module or function, optional
        The activation function to be applied after linearly combining
        features. Must be a callable that accepts a tensor as sole
        argument, like a module from ``torch.nn`` or a function from
        `torch.nn.functional``, depending on whether it needs to be further
        parameterized or not. Defaults to ``identity``, resulting in no
        non-linear activation whatsoever.
    **kwargs
        Keyword arguments are passed on to the linear layer.

    """

    def __init__(
            self,
            mod_dim: int,
            n_features: int,
            activate: Module | Functional = identity,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.n_features = n_features
        self.activate = activate
        self.kwargs = kwargs
        self.mix = ptn.Linear(n_features * mod_dim, mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass for combining multiple stacked feature vectors.

        Parameters
        ----------
        inp: Tensor
            The size of the next-to-last last dimension of the input tensor is
            expected to match the `n_features` provided at instantiation.
            The last dimension (of size `mod_dim`) is expected to contain the
            features vectors themselves.

        Returns
        -------
        Tensor
            The output tensor has one fewer dimensions than the input.
            The next-to-last dimension is dropped and the size of the last
            dimension is once again `mod_dim`.

        """
        return self.activate(self.mix(inp.flatten(start_dim=-2)))

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.mix.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            n_features: int | None = None,
            activate: Module | Functional | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Size of the feature space. The input tensor is expected to be of
            that size in its last dimension and the output will again have this
            size in its last dimension. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        n_features: int, optional
            The number of features to combine. Must be equal to the size of the
            next-to-last dimension of the input tensor. Overwrites `n_features`
            of the current instance if given. Defaults to ``None``.
        activate: Module or function, optional
            The activation function to be applied after linearly combining
            features. Must be a callable that accepts a tensor as sole
            argument, like a module from ``torch.nn`` or a function from
            `torch.nn.functional``, depending on whether it needs to be further
            parameterized or not. Overwrites the `activate` of the current
            instance if given. Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layer together.

        Returns
        -------
        ActivatedConcatMixer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.n_features if n_features is None else n_features,
            self.activate if activate is None else activate,
            **(self.kwargs | kwargs)
        )



import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any
from pathlib import Path
import shutil
from google.cloud import storage as gcs
from ...misc import ArgRepr


class GcsDir2LocalDir(ArgRepr):
    """Download files from Google Cloud Storage to local directory.

    Parameters
    ----------
    project: str
        Project where the `bucket` and blobs reside.
    bucket: str
        Bucket where the blobs reside.
    prefix: str, optional
        The prefix of the blobs to download. Since it (or part of it)
        can also be provided later, when the callable instance is called, it
        is optional here. Defaults to an empty string.
    base_dir: str, optional
        Absolute path to a base directory on the local filesystem.
        Defaults to "/tmp".
    overwrite: bool, optional
        Whether to silently overwrite local destination directory. Defaults
        to ``False``, which will raise an exception if it already exists.
    skip: bool, optional
        Whether to simply return the list of files found in the local
        destination directory, if it and any exists. Defaults to ``False``.
    n_threads: int, optional
        Maximum number of blobs to download in parallel. Defaults to 16.
    chunk_size: int, optional
        Chunk size to read from Google Cloud Storage in one API call in MiB.
        Defaults to 10 MiB.
    **kwargs
        Additional keyword arguments are passed to the constructor of the
        Google Storage ``Client`` (see `documentation <https://cloud.google.
        com/python/docs/reference/storage/latest/google.cloud.storage.
        client.Client#parameters>`__ for options).

    Notes
    -----
    Blobs that have any more forward slashes in their name than the
    instantiation and call prefixes combined, i.e., that reside in virtual
    "subdirectories", are ignored and are not downloaded.

    """

    __thread = threading.local()

    def __init__(
            self,
            project: str,
            bucket: str,
            prefix: str = '',
            base_dir: str = '/tmp',
            overwrite: bool = False,
            skip: bool = False,
            n_threads: int = 16,
            chunk_size: int = 10,
            **kwargs: Any
    ) -> None:
        self.project = project.strip(' /.')
        self.bucket = bucket.strip(' /.')
        self.prefix = prefix.strip(' ./') + '/' if prefix.strip(' ./') else ''
        self.base_dir = '/' + base_dir.strip(' /')
        self.overwrite = overwrite
        self.skip = skip
        self.n_threads = n_threads
        self.chunk_size = chunk_size
        self.kwargs = kwargs
        super().__init__(
            self.project,
            self.bucket,
            self.prefix,
            self.base_dir,
            self.overwrite,
            self.skip,
            self.n_threads,
            self.chunk_size,
            **kwargs
        )

    @property
    def chunk_bytes(self) -> int:
        """Bytes to read from Google Cloud Storage in one API call."""
        in_bytes = self.chunk_size * 1024 * 1024
        in_multiples_of_256kb = int(in_bytes // (256 * 1024))
        return in_multiples_of_256kb * 256 * 1024

    def __call__(self, prefix: str = '') -> list[str]:
        """Download files from Google Cloud Storage to local drive.

        Parameters
        ----------
        prefix: str, optional
            The prefix of the files to download. If given here, it will
            be appended to the `prefix` given at instantiation time.
            Defaults to an empty string.

        Returns
        -------
        list
            A list of the fully resolved file names on the local drive.

        Raises
        ------
        FileExistsError
            If `overwrite` is ``False`` and download into an existing folder
            was attempted.

        """
        prefix = prefix.strip(' ./') + '/' if prefix.strip(' ./') else ''
        local = self.prefix + prefix

        files = self.__files_from(local)

        if not files:
            client = gcs.Client(self.project, **self.kwargs)
            blobs = client.list_blobs(self.bucket, prefix=local or None)
            with ThreadPoolExecutor(
                    self.n_threads,
                    initializer=self.__initializer
            ) as pool:
                downloads = as_completed(
                    pool.submit(self.__download, blob.name, local)
                    for blob in blobs
                    if blob.name.count('/') == local.count('/')
                )
                files = [download.result() for download in downloads]

        return files

    def __files_from(self, local: str) -> list[str]:
        """Return files from local directory if configured adn present."""
        path = Path(self.base_dir) / local

        if path.exists() and self.overwrite:
            resolved = str(path.resolve())
            shutil.rmtree(resolved) if path.is_dir() else path.unlink()

        if path.exists() and path.is_dir() and self.skip:
            return [
                str(obj.resolve())
                for obj in path.iterdir()
                if obj.is_file()
            ]

        if path.exists() and path.is_dir() and not any(path.iterdir()):
            return []

        path.mkdir(parents=True)
        return []

    def __initializer(self) -> None:
        """Each thread needs its own client as they cannot be pickled."""
        self.__thread.client = gcs.Client(self.project, **self.kwargs)
        self.__thread.bucket = self.__thread.client.get_bucket(self.bucket)

    def __download(self, name: str, local: str) -> str:
        """Download a single file from Google Cloud Storage to local."""
        blob = self.__thread.bucket.get_blob(name)
        blob.chunk_size = self.chunk_bytes
        file = blob.name.split('/')[-1]
        path = Path(self.base_dir) / local / file
        with path.open('wb') as stream:
            blob.download_to_file(stream, raw_download=True)
        return str(path.resolve())



from collections.abc import Callable, Iterable
from ..misc import ArgRepr
from .exceptions import SplitError


class Split[S, T](ArgRepr):
    """Split an iterable of objects into two according to a decision criterion.

    Upon subclassing and/or instantiation, type annotation with the type of the
    elements in the iterator to be filtered and the return type of `wrapper`
    is recommended.

    Parameters
    ----------
    criterion: callable, optional
        The `condition` is called with one object at a time and must return a
        boolean value. Objects on which it evaluates to ``True`` are sorted
        into one container and those where it evaluates to ``False`` end up
        in a second container. Defaults to ``None``, which returns the inherent
        truth values of the object in the iterable.
    wrapper: type or callable, optional
        If not given, an attempt will be made to determine the type of the
        container the callable instance was called with. Whether inferred or
        explicitly given, `wrapper` will be called twice, once with a list of
        a list of the elements that evaluated to ``True`` and once with those
        that evaluated to ``False``.

    """

    def __init__(
            self,
            criterion: Callable[[S], bool] | None = None,
            wrapper: type[T] | Callable[[list[S]], T] | None = None
    ) -> None:
        super().__init__(criterion, wrapper)
        self.criterion = criterion
        self.wrapper = wrapper

    def __call__(self, iterable: Iterable[S]) -> tuple[T, T]:
        """Split sequence into two according to cached decision criterion.

        Parameters
        ----------
        iterable: Iterable
            Objects to be split according to the cached (boolean) `criterion`.

        Returns
        -------
        tuple
            Two sequences of the same type as the input sequence or, if a
            `wrapper` was specified, of that type. Objects where the
            `criterion` evaluated to ``True`` are in the first return
            sequence and those evaluating to ``False`` are in the second.

        Raises
        ------
        SplitError
            If calling the `criterion` on any element of `iterable` raises
            an exception or if wrapping the results leads to an exception.

        """
        criterion = bool if self.criterion is None else self.criterion
        true = []
        false = []
        for i, element in enumerate(iterable):
            try:
                criterion_is_fulfilled = criterion(element)
            except Exception as error:
                msg = '\n{} calling\n{}\non element #{}:\n{}\n{}'
                name = self._name(criterion)
                err_cls = error.__class__.__name__
                fmt = msg.format(err_cls, name, i, element, error)
                raise SplitError(fmt) from error
            if criterion_is_fulfilled:
                true.append(element)
            else:
                false.append(element)
        wrap = iterable.__class__ if self.wrapper is None else self.wrapper
        try:
            true = wrap(true)
            false = wrap(false)
        except Exception as error:
            msg = '\n{} calling wrapper\n{}\non split results:\n{}'
            name = self._name(wrap)
            err_cls = error.__class__.__name__
            raise SplitError(msg.format(err_cls, name, error)) from error
        return true, false



"""Flexibly project your features into embedding space.

The first step in many modern neural-network architectures is to transform
input features into vectors in an embedding space with a certain number of
dimensions, the "model dimension" (or overall "bus width" of the model). This
subpackage provides several ways to do that for both numerical and categorical
features so that, when combined, all are treated on equal footing.

"""

from .activated import ActivatedEmbedder
from .gated import GatedEmbedder
from .gated_residual import GatedResidualEmbedder
from .numerical import NumericalEmbedder
from .categorical import CategoricalEmbedder
from .feature import FeatureEmbedder

__all__ = [
    'ActivatedEmbedder',
    'GatedEmbedder',
    'GatedResidualEmbedder',
    'NumericalEmbedder',
    'CategoricalEmbedder',
    'FeatureEmbedder'
]



from abc import ABC, abstractmethod


class CustomField(ABC):
    """Abstract base class for defining custom field types in a JsonObject."""

    @abstractmethod
    def __init__(self, arg) -> None:
        """Will be called with exactly one argument as a custom field type."""
        ...

    @property
    @abstractmethod
    def as_json(self):
        """How should instances be represented in a JSON-serializable dict?"""
        ...

    @property
    @abstractmethod
    def as_dtype(self):
        """How should instances appear in cells of a pandas Series?"""
        ...



"""Combine feature embedding through weighted sums.

Depending on whether these weights are learnable themselves and on whether
they depend on the input features or not, some type of feature importance
can be provided.

"""

from .constant import ConstantSumMixer
from .variable import VariableSumMixer
from .activated import ActivatedSumMixer
from .gated import GatedSumMixer
from .gated_residual import GatedResidualSumMixer

__all__ = [
    'ConstantSumMixer',
    'VariableSumMixer',
    'ActivatedSumMixer',
    'GatedSumMixer',
    'GatedResidualSumMixer'
]



from typing import Any, Self
import datetime as dt
import pandas as pd
from pandas import Timestamp, Timedelta

Date = str | dt.date | dt.datetime | Timestamp
Delta = dt.timedelta | Timedelta


class FlexiDate:
    """Flexible wrapper around python's own ``datetime.date`` object.

    Parameters
    ----------
    date
        Can be an ISO string of a date or datetime, a ``datetime.date`` or
        a ``datetime.datetime`` object, or a ``pandas.Timestamp``.

    """

    def __init__(self, date: Date | Self) -> None:
        self.as_date = dt.datetime.fromisoformat(str(date)).date()

    def __getattr__(self, attribute: str) -> Any:
        return getattr(self.as_date, attribute)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{str(self)}')"

    def __str__(self) -> str:
        return str(self.as_date)

    def __lt__(self, other: Date) -> bool:
        return self.as_date < self.__class__(other).as_date

    def __le__(self, other: Date) -> bool:
        return self.as_date <= self.__class__(other).as_date

    def __gt__(self, other: Date) -> bool:
        return self.as_date > self.__class__(other).as_date

    def __ge__(self, other: Date) -> bool:
        return self.as_date >= self.__class__(other).as_date

    def __eq__(self, other: Date) -> bool:
        return self.__class__(other).as_date == self.as_date

    def __ne__(self, other: Date) -> bool:
        return self.__class__(other).as_date != self.as_date

    def __add__(self, other: Delta) -> Self:
        return self.__class__(self.as_date + other)

    def __radd__(self, other: Delta) -> Self:
        return self.__class__(other + self.as_date)

    def __sub__(self, other: Delta) -> Self:
        return self.__class__(self.as_date - other)

    def __hash__(self) -> int:
        return self.as_date.__hash__()

    @property
    def as_datetime(self) -> dt.datetime:
        """Date as a datetime object."""
        return dt.datetime(
            self.as_date.year,
            self.as_date.month,
            self.as_date.day
        )

    @property
    def as_json(self) -> str:
        """Representation to appear in a JSON."""
        return str(self)

    @property
    def as_dtype(self) -> Timestamp:
        """Representation in a pandas DataFrame."""
        return pd.to_datetime(self.as_date)



from typing import Any
from pathlib import Path
import pandas as pd
from pandas import DataFrame
from ..misc import ArgRepr


class ParquetReader(ArgRepr):
    """Light wrapper around the top-level ``read_parquet`` pandas function.

    Parameters
    ----------
    path: str, optional
        Directory under which the parquet files are located or full path to
        the parquet file. If not fully specified here, the path must be
        completed on calling the instance. Defaults to the current working
        directory of the python interpreter.
    **kwargs
        Keyword arguments passed on to the ``read_parquet`` function call.

    """

    def __init__(self, path: str = '', **kwargs: Any) -> None:
        self.path = path.strip()
        self.kwargs = kwargs
        super().__init__(self.path, **kwargs)

    def __call__(self, path: str = '') -> DataFrame:
        """Read one or more parquet file(s) into pandas DataFrame.

        Parameters
        ----------
        path: str, optional
            Path (including file name) relative to the `path` specified at
            instantiation. Defaults to an empty string, which results in an
            unchanged `path` on concatenation.

        Returns
        -------
        DataFrame

        """
        path = Path(self.path) / path.strip(' /')
        return pd.read_parquet(path, **self.kwargs).reset_index(drop=True)



"""Tools for composing functions and other callable objects into workflows.

Once you have broken down your workflow into small, modular steps, the tools in
this package allow you to parameterize them with everything that is known at
program start, and compose them into arbitrarily complex execution graphs that
are easy to maintain and to extend.

Important
---------
In order for everything to work smoothly, there is only one thing you will have
to keep in mind when writing the composable building blocks of your workflow.
If a function (or other callable object) is to return *nothing*, that is, when
it is a *side effect*, don't just omit the ``return`` statement, but explicitly
return an *empty tuple*!

"""

from .filter import Filter
from .fork import Fork
from .map import Map
from .partial import Partial
from .curry import Curry
from .pipe import Pipe
from .reduce import Reduce
from .route import Route
from .split import Split
from .sum import Sum
from .safe import Safe
from .exceptions import SafeError
from .misc import apply, unit, identity

__all__ = [
    'Filter',
    'Fork',
    'Map',
    'Partial',
    'Curry',
    'Pipe',
    'Reduce',
    'Route',
    'Split',
    'Sum',
    'Safe',
    'SafeError',
    'apply',
    'unit',
    'identity'
]

# ToDo: Add Fallback and Retry
# ToDo: Add lazy sub-package



from typing import Self
import torch as pt
from ..types import Tensor, Module
from ..exceptions import EmbeddingError
from .numerical import NumericalEmbedder
from .categorical import CategoricalEmbedder


class FeatureEmbedder(Module):
    """Jointly embed numerical and categorical features into stacked vectors.

    Given a float tensor where both, numerical and categorical features
    appear (one before the other in the last dimension), instances of this
    class treat them on equal footing and produce stacked embedding vectors
    for all of them.

    Parameters
    ----------
    embed_num: NumericalEmbedder
        A fully configured instance of ``NumericalEmbedder``.
    embed_cat: CategoricalEmbedder
        A fully configured instance of ``CategoricalEmbedder``.

    Raises
    ------
    EmbeddingError
        If the embedding dimension of the numerical and the categorical
        embedders do not match.

    """

    def __init__(
            self,
            embed_num: NumericalEmbedder,
            embed_cat: CategoricalEmbedder
    ) -> None:
        super().__init__()
        if embed_num.mod_dim != embed_cat.mod_dim:
            msg = (f'Numerical ({embed_num.mod_dim}) and categorical ('
                   f'{embed_cat.mod_dim}) embedding dimensions must match!')
            raise EmbeddingError(msg)
        self.embed_num = embed_num
        self.embed_cat = embed_cat

    @property
    def n_num(self) -> int:
        """Number of numerical features."""
        return self.embed_num.n_features

    @property
    def n_cat(self) -> int:
        """Number of categorical features."""
        return self.embed_cat.n_features

    @property
    def n_features(self) -> int:
        """Total number of features."""
        return self.n_num + self.n_cat

    def forward(self, inp: Tensor) -> Tensor:
        """Forward pass for numerical and categorical feature embeddings.

        Parameters
        ----------
        inp: Tensor
            Input tensor of must be of dtype ``float``. The last dimension is
            expected to contain first the values of all numerical features,
            followed by those of the categorical features.

        Returns
        -------
        Tensor
            The output tensor has one more dimension of size `mod_dim` added
            after the last dimension (with a size equal to the total number of
            features) than the `inp`, containing the stacked embeddings, first
            those of the numerical and then those of the categorical features.

        """
        return pt.cat([
            self.embed_num(inp[..., :self.n_num]),
            self.embed_cat(inp[..., self.n_num:].long())
        ], dim=-2)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.embed_num.reset_parameters()
        self.embed_cat.reset_parameters()

    def new(
            self,
            embed_num: NumericalEmbedder | None = None,
            embed_cat: CategoricalEmbedder | None = None
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        embed_num: NumericalEmbedder, optional
            Overwrites the `embed_num` of the current instance if given.
            Defaults to ``None``.
        embed_cat: CategoricalEmbedder, optional
            Overwrites the `embed_cat` of the current instance if given.
            Defaults to ``None``.

        Returns
        -------
        FeatureEmbedder
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.embed_num.new() if embed_num is None else embed_num,
            self.embed_cat.new() if embed_cat is None else embed_cat
        )



from typing import Any, overload
from collections.abc import Callable


def apply[**P, T](call: type[T] | Callable[P, T], *args: P.args) -> T:
    """Call a callable object with the specified arguments.

    Parameters
    ----------
    call: callable
        The callable object to call with `args`.
    *args
        The arguments to call `call` with.

    Returns
    -------
    object
        Whatever `call` returns.

    """
    return call(*args)


def unit(*_: Any) -> tuple[()]:
    """Accepts any number of arguments but always returns and empty tuple.

    Returns
    -------
    tuple
        Empty tuple.

    """
    return ()


@overload
def identity() -> tuple[()]:
    ...


@overload
def identity[T](arg: T) -> T:
    ...


@overload
def identity[T, *Ts](arg: T, *args: *Ts) -> tuple[T, *Ts]:
    ...


def identity(*args):
    """Pass through any number of argument(s) doing nothing.

    Parameters
    ----------
    *args
        Argument(s) to pass through.

    Returns
    -------
    object or tuple
        Called with one argument, this argument is simply passed through.
        When called with more than one argument, the arguments tuple returned.
        When called with no argument, an empty tuple is returned.

    """
    return args[0] if len(args) == 1 else args



"""Convenient classes (and functions) that do not fit any other category."""

from typing import Any, Self
import torch.nn as ptn
from .types import Tensor, Module, Tensors, Tensors2T

__all__ = [
    'identity',
    'Identity',
    'Finalizer',
    'NegativeBinomialFinalizer'
]


def identity(tensor: Tensor, **_: Any) -> Tensor:
    """Simply pass through the argument and ignore keyword arguments.

    This is a placeholder for instances where a default function is required.

    Parameters
    ----------
    tensor: Tensor
        Any argument (typically a tensor) to be passed straight through.

    Returns
    -------
    Tensor
        The tensor passed in as argument.

    """
    return tensor


class Identity(Module):
    """PyTorch module that passes a tensor right through, doing nothing.

    This is a placeholder for instances where a default module is required.
    Providing any number of (keyword) arguments on instantiation is permitted,
    but they will be ignored.

    """

    def __init__(self, *_, **__) -> None:
        super().__init__()

    def forward(self, tensor: Tensor, **_: Any) -> Tensor:
        """Simply pass through the argument and ignore keyword arguments.

        Parameters
        ----------
        tensor: Tensor
            Any argument (typically a tensor) to be passed straight through.

        Returns
        -------
        Tensor
            The tensor passed in as argument.

        """
        return tensor

    def reset_parameters(self) -> None:
        """Does nothing because there are no internal parameters to reset."""

    def new(self, *_, **__) -> Self:
        """Return a fresh, new instance.

        Providing any number of (keyword) arguments is permitted, but they will
        be ignored.

        Returns
        -------
        Identity
            A fresh, new instance of itself.

        """
        return self.__class__()


class Finalizer(Module):
    """Extract one or more numbers from the final layer of a neural network.

    Instances of this class serve as a convenient final layer in any neural
    network, no matter if it is for regression, for classification, for
    multiple targets, or if you predict the parameters of a probability
    distribution. The last activations of your network are going to be passed
    through as many linear layers as you need outputs, each passed through
    their own (and potentially different) non-linearity to give you the desired
    number of outputs and the desired value range for each output.

    Parameters
    ----------
    mod_dim: int
        The size of the last dimension of the input tensor, essentially the
        "width" of the neural network before it is to be collapsed to the
        final output.
    *activations: Module
        Specify as many activations (instances of PyTorch ``Module``) as you
        want outputs (e.g., ``Sigmoid()`` for binary classification,
        ``Softplus()`` for strictly positive regression targets, etc.). For
        unbounded regression targets, where you want no activation function at
        all, use ``Identity()``.
    **kwargs
        Keyword arguments are passed on to all linear layers.

    See Also
    --------
    Identity

    """

    def __init__(
            self,
            mod_dim: int,
            *activations: Module,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.activations: tuple[Module, ...] = activations
        self.kwargs = kwargs
        self.finalize = ptn.ModuleList(
            ptn.Sequential(ptn.Linear(mod_dim, 1, **kwargs), activate)
            for activate in activations
        )

    @property
    def n_out(self) -> int:
        """Number of final outputs."""
        return len(self.activations)

    def forward(self, inp: Tensor) -> Tensors:
        """Forward pass for extracting outputs from the final hidden layer.

        Parameters
        ----------
        inp: Tensor
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        tuple
            As many tensors are returned as there were `activations` specified.
            Each tensor has the same dimension and size, viz., that of the
            `inp` with the size of the last dimension shrunk to 1.

        """
        return tuple(finalize(inp) for finalize in self.finalize)

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        for finalize in self.finalize:
            finalize[0].reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            *activations: Module,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            The size of the last dimension of the input tensor. Overwrites the
            `mod_dim` of the current instance if given. Defaults to ``None``.
        *activations: Module
            Activation functions replace the ones in the current instance if
            any are given. If none are given, the new instance will have the
            same as the present instance.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layers together.

        Returns
        -------
        Finalizer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            *(activations or self.activations),
            **(self.kwargs | kwargs)
        )


class NegativeBinomialFinalizer(Module):
    """Consistent mean and standard deviation for over-dispersed counts.

    When regressing potentially over-dispersed counts data, you might want to
    use the negative log-likelihood as loss function. However, this is only
    defined if the variance is greater than the mean. Following D. Salinas
    `et al.`, [1]_ this can be achieved by extracting a (positive) mean value
    and a (positive) scale factor from the last hidden layer of your network,
    and letting the variance be the sum of mean and the scaled square of the
    mean.

    Parameters
    ----------
    mod_dim: int
        Size of the feature space. The input tensor is expected to be of that
        size in its last dimension.
    beta: float, optional
        Scaling parameter ot the `Softplus <https://pytorch.org/docs/stable/
        generated/torch.nn.Softplus.html#torch.nn.Softplus>`__ activation
        function. Defaults to 1.
    threshold: float, optional
        The `Softplus <https://pytorch.org/docs/stable/generated/
        torch.nn.Softplus.html#torch.nn.Softplus>`__ activation is approximated
        as a linear function for values greater than this. Defaults to 20.
    **kwargs
        Keyword arguments are passed on to the linear layers.

    See Also
    --------
    swak.pt.losses.NegativeBinomialLoss
    swak.pt.dists.MuSigmaNegativeBinomial

    References
    ----------
    .. [1] D. Salinas, V. Flunkert, and J. Gasthaus, and T. Pfister, `DeepAR:
           Probabilistic Forecasting with Autoregressive Recurrent Networks`,
           `arXiv:1704.04110v3 <https://arxiv.org/pdf/1704.04110>`__ (2019).

    """

    def __init__(
            self,
            mod_dim: int,
            beta: float = 1.0,
            threshold: float = 20.0,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        self.beta = beta
        self.threshold = threshold
        self.kwargs = kwargs
        self.mu = ptn.Linear(mod_dim, 1, **kwargs)
        self.alpha = ptn.Linear(mod_dim, 1, **kwargs)
        self.activate = ptn.Softplus(beta, threshold)

    def forward(self, inp: Tensor) -> Tensors2T:
        """Forward pass for generating mean and matching standard deviation.

        Parameters
        ----------
        inp: Tensor
            The activations after the last hidden layer in your network.
            The size of the last dimension is expected to be ``mod_dim``.

        Returns
        -------
        tuple
            A tensor with the predicted mean values and a tensor with the
            predicted standard deviations that are guaranteed to be greater
            or equal to the square root of the mean. Both have the same
            dimension and size, viz., that of the `inp` with the size of the
            last dimension shrunk to 1.

        """
        mu = self.activate(self.mu(inp))
        alpha = self.activate(self.alpha(inp))
        return mu, (mu * (1.0 + mu * alpha)).sqrt()

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.mu.reset_parameters()
        self.alpha.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            beta: float | None = None,
            threshold: float | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            The size of the last dimension of the input tensor. Overwrites the
            `mod_dim` of the current instance if given. Defaults to ``None``.
        beta: float, optional
            Scaling parameter ot the ``Softplus`` activation function.
            Overwrites the `beta` of the current instance if given.
            Defaults to ``None``.
        threshold: float, optional
            The ``Softplus`` activation is approximated as a linear function
            for values greater than this. Overwrites the `threshold` of the
            current instance if given. Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layers together.

        Returns
        -------
        NegativeBinomialFinalizer
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.beta if beta is None else beta,
            self.threshold if threshold is None else threshold,
            **(self.kwargs | kwargs)
        )
